# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

import logging
import os
import re

from jinja2 import Environment
from material.utilities.filter import FileFilter
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.exceptions import PluginError
from mkdocs.plugins import BasePlugin, event_priority
from mkdocs.structure.pages import Page
from mkdocs.utils.templates import TemplateContext

from .config import TagsConfig
from .renderer import Renderer
from .structure.listing.manager import ListingManager
from .structure.mapping.manager import MappingManager
from .structure.mapping.storage import MappingStorage

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

class TagsPlugin(BasePlugin[TagsConfig]):
    """
    A tags plugin.

    This plugin collects tags from the front matter of pages, and builds a tag
    structure from them. The tag structure can be used to render listings on
    pages, or to just create a site-wide tags index and export all tags and
    mappings to a JSON file for consumption in another project.
    """

    supports_multiple_instances = True
    """
    This plugin supports multiple instances.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the plugin.
        """
        super().__init__(*args, **kwargs)

        # Initialize incremental builds
        self.is_serve = False

        # Initialize mapping and listing managers
        self.mappings = None
        self.listings = None

    # -------------------------------------------------------------------------

    mappings: MappingManager
    """
    Mapping manager.
    """

    listings: ListingManager
    """
    Listing manager.
    """

    filter: FileFilter
    """
    File filter.
    """

    # -------------------------------------------------------------------------

    def on_startup(self, *, command, **kwargs) -> None:
        """
        Determine whether we're serving the site.

        Arguments:
            command: The command that is being executed.
            dirty: Whether dirty builds are enabled.
        """
        self.is_serve = command == "serve"

    def on_config(self, config: MkDocsConfig) -> None:
        """
        Create mapping and listing managers.
        """

        # Retrieve toc depth, so we know the maximum level at which we can add
        # items to the table of contents - Python Markdown allows to set the
        # toc depth as a range, e.g. `2-6`, so we need to account for that as
        # well. We need this information for generating listings.
        depth = config.mdx_configs.get("toc", {}).get("toc_depth", 6)
        if not isinstance(depth, int) and "-" in depth:
            _, depth = depth.split("-")

        # Initialize mapping and listing managers
        self.mappings = MappingManager(self.config)
        self.listings = ListingManager(self.config, int(depth))

        # Initialize file filter - the file filter is used to include or exclude
        # entire subsections of the documentation, allowing for using multiple
        # instances of the plugin alongside each other. This can be necessary
        # when creating multiple, potentially conflicting listings.
        self.filter = FileFilter(self.config.filters)

        # Ensure presence of attribute lists extension
        for extension in config.markdown_extensions:
            if isinstance(extension, str) and extension.endswith("attr_list"):
                break
        else:
            config.markdown_extensions.append("attr_list")

        # If the author only wants to extract and export mappings, we allow to
        # disable the rendering of all tags and listings with a single setting
        if self.config.export_only:
            self.config.tags = False
            self.config.listings = False

        # By default, shadow tags are rendered when the documentation is served,
        # but not when it is built, for a better user experience
        if self.is_serve and self.config.shadow_on_serve:
            self.config.shadow = True

    @event_priority(-50)
    def on_page_markdown(
        self, markdown: str, *, page: Page, config: MkDocsConfig, **kwargs
    ) -> str:
        """
        Collect tags and listings from page.

        Priority: -50 (run later)

        Arguments:
            markdown: The page's Markdown.
            page: The page.
            config: The MkDocs configuration.

        Returns:
            The page's Markdown with injection points.
        """
        if not self.config.enabled:
            return

        # Skip if page should not be considered
        if not self.filter(page.file):
            return

        # Handle deprecation of `tags_file` setting
        if self.config.tags_file:
            markdown = self._handle_deprecated_tags_file(page, markdown)

        # Handle deprecation of `tags_extra_files` setting
        if self.config.tags_extra_files:
            markdown = self._handle_deprecated_tags_extra_files(page, markdown)

        # Collect tags from page
        try:
            self.mappings.add(page, markdown)

        # Raise exception if tags could not be read
        except Exception as e:
            docs = os.path.relpath(config.docs_dir)
            path = os.path.relpath(page.file.abs_src_path, docs)
            raise PluginError(
                    f"Error reading tags of page '{path}' in '{docs}':\n"
                    f"{e}"
                )

        # Collect listings from page
        return self.listings.add(page, markdown)

    @event_priority(100)
    def on_env(
        self, env: Environment, *, config: MkDocsConfig, **kwargs
    ) -> None:
        """
        Populate listings.

        Priority: 100 (run earliest)

        Arguments:
            env: The Jinja environment.
            config: The MkDocs configuration.
        """
        if not self.config.enabled:
            return

        # Populate and render all listings
        self.listings.populate_all(self.mappings, Renderer(env, config))

        # Export mappings to file, if enabled
        if self.config.export:
            path = os.path.join(config.site_dir, self.config.export_file)
            path = os.path.normpath(path)

            # Serialize mappings and save to file
            storage = MappingStorage(self.config)
            storage.save(path, self.mappings)

    def on_page_context(
        self, context: TemplateContext, *, page: Page, **kwargs
    ) -> None:
        """
        Add tag references to page context.

        Arguments:
            context: The template context.
            page: The page.
        """
        if not self.config.enabled:
            return

        # Skip if page should not be considered
        if not self.filter(page.file):
            return

        # Skip if tags should not be built
        if not self.config.tags:
            return

        # Retrieve tags references for page
        mapping = self.mappings.get(page)
        if mapping:
            tags = self.config.tags_name_variable
            if tags not in context:
                context[tags] = list(self.listings & mapping)

    # -------------------------------------------------------------------------

    def _handle_deprecated_tags_file(
        self, page: Page, markdown: str
    ) -> str:
        """
        Handle deprecation of `tags_file` setting.

        Arguments:
            page: The page.
        """
        directive = self.config.listings_directive
        if page.file.src_uri != self.config.tags_file:
            return markdown

        # Try to find the legacy tags marker and replace with directive
        if "[TAGS]" in markdown:
            markdown = markdown.replace(
                "[TAGS]", f"<!-- {directive} -->"
            )

        # Try to find the directive and add it if not present
        pattern = r"<!--\s+{directive}".format(directive = directive)
        if not re.search(pattern, markdown):
            markdown += f"\n<!-- {directive} -->"

        # Return markdown
        return markdown

    def _handle_deprecated_tags_extra_files(
        self, page: Page, markdown: str
    ) -> str:
        """
        Handle deprecation of `tags_extra_files` setting.

        Arguments:
            page: The page.
        """
        directive = self.config.listings_directive
        if page.file.src_uri not in self.config.tags_extra_files:
            return markdown

        # Compute tags to render on page
        tags = self.config.tags_extra_files[page.file.src_uri]
        if tags:
            directive += f" {{ include: [{', '.join(tags)}] }}"

        # Try to find the legacy tags marker and replace with directive
        if "[TAGS]" in markdown:
            markdown = markdown.replace(
                "[TAGS]", f"<!-- {directive} -->"
            )

        # Try to find the directive and add it if not present
        pattern = r"<!--\s+{directive}".format(directive = re.escape(directive))
        if not re.search(pattern, markdown):
            markdown += f"\n<!-- {directive} -->"

        # Return markdown
        return markdown

# -----------------------------------------------------------------------------
# Data
# -----------------------------------------------------------------------------

# Set up logging
log = logging.getLogger("mkdocs.material.plugins.tags")



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import os
import posixpath

from mergedeep import Strategy, merge
from mkdocs.exceptions import PluginError
from mkdocs.structure.files import InclusionLevel
from mkdocs.plugins import BasePlugin, event_priority
from yaml import SafeLoader, load

from .config import MetaConfig

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Meta plugin
class MetaPlugin(BasePlugin[MetaConfig]):

    # Construct metadata mapping
    def on_files(self, files, *, config):
        if not self.config.enabled:
            return

        # Initialize mapping
        self.meta = {}

        # Resolve and load meta files in docs directory
        docs = os.path.relpath(config.docs_dir)
        for file in files:
            name = posixpath.basename(file.src_uri)
            if not name == self.config.meta_file:
                continue

            # Exclude meta file from site directory - explicitly excluding the
            # meta file allows the author to use a file name without '.' prefix
            file.inclusion = InclusionLevel.EXCLUDED

            # Open file and parse as YAML
            with open(file.abs_src_path, encoding = "utf-8-sig") as f:
                path = file.src_path
                try:
                    self.meta[path] = load(f, SafeLoader)

                # The meta file could not be loaded because of a syntax error,
                # which we display to the author with a nice error message
                except Exception as e:
                    raise PluginError(
                        f"Error reading meta file '{path}' in '{docs}':\n"
                        f"{e}"
                    )

    # Set metadata for page, if applicable (run earlier)
    @event_priority(50)
    def on_page_markdown(self, markdown, *, page, config, files):
        if not self.config.enabled:
            return

        # Start with a clean state, as we first need to apply all meta files
        # that are relevant to the current page, and then merge the page meta
        # on top of that to ensure that the page meta always takes precedence
        # over meta files - see https://t.ly/kvCRn
        meta = {}

        # Merge matching meta files in level-order
        strategy = Strategy.TYPESAFE_ADDITIVE
        for path, defaults in self.meta.items():
            if not page.file.src_path.startswith(os.path.dirname(path)):
                continue

            # Skip if meta file was already merged - this happens in case of
            # blog posts, as they need to be merged when posts are constructed,
            # which is why we need to keep track of which meta files are applied
            # to what pages using the `__extends` key.
            page.meta.setdefault("__extends", [])
            if path in page.meta["__extends"]:
                continue

            # Try to merge metadata
            try:
                merge(meta, defaults, strategy = strategy)
                page.meta["__extends"].append(path)

            # Merging the metadata with the given strategy resulted in an error,
            # which we display to the author with a nice error message
            except Exception as e:
                docs = os.path.relpath(config.docs_dir)
                raise PluginError(
                    f"Error merging meta file '{path}' in '{docs}':\n"
                    f"{e}"
                )

        # Ensure page metadata is merged last, so the author can override any
        # defaults from the meta files, or even remove them entirely
        page.meta = merge(meta, page.meta, strategy = strategy)

# -----------------------------------------------------------------------------
# Data
# -----------------------------------------------------------------------------

# Set up logging
log = logging.getLogger("mkdocs.material.meta")



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

import posixpath

from collections.abc import Iterator
from material.plugins.tags.structure.mapping import Mapping
from material.plugins.tags.structure.tag import Tag
from mkdocs.structure.pages import Page

from .config import ListingConfig
from .tree import ListingTree

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

class Listing:
    """
    A listing of tags.

    Listings can be included on a page with the `<!-- material/tags [args] -->`
    directive. The arguments are passed to a YAML parser, and are expected to
    either be an inline listing configuration, or an identifier that points to
    a listing configuration in the `listings_map` setting in `mkdocs.yml`.

    Example 1: use inline listing configuration

    ``` markdown
    <!-- material/tags { include: [foo, bar] } -->
    ```

    Example 2: use listing configuration from `listings_map.qux`

    ``` markdown
    <!-- material/tags qux -->
    ```

    If no arguments are given, the default listing configuration is used,
    rendering all tags and mappings. In case you're using multiple instances
    of the tags plugin, you can use the `listings_directive` setting to change
    the directive name.
    """

    def __init__(self, page: Page, id: str, config: ListingConfig):
        """
        Initialize the listing.

        Arguments:
            page: The page the listing is embedded in.
            id: The listing identifier.
            config: The listing configuration.
        """
        self.page = page
        self.id = id
        self.config = config
        self.tags = {}

    def __repr__(self) -> str:
        """
        Return a printable representation of the listing.

        Returns:
            Printable representation.
        """
        return f"Listing({repr(self.page)})"

    def __hash__(self) -> int:
        """
        Return the hash of the listing.

        Returns:
            The hash.
        """
        return hash(self.id)

    def __iter__(self) -> Iterator[ListingTree]:
        """
        Iterate over the listing in pre-order.

        Yields:
            The current listing tree.
        """
        stack = list(reversed(self.tags.values()))
        while stack:
            tree = stack.pop()
            yield tree

            # Visit subtrees in reverse, so pre-order is preserved
            stack += reversed([*tree])

    def __and__(self, mapping: Mapping) -> Iterator[Tag]:
        """
        Iterate over the tags of a mapping featured in the listing.

        When hierarchical tags are used, the set of tags is expanded to include
        all parent tags, but only for the inclusion check. The returned tags are
        always the actual tags (= leaves) of the mapping. This is done to avoid
        duplicate entries in the listing.

        If a mapping features one of the tags excluded from the listing, the
        entire mapping is excluded from the listing. Additionally, if a listing
        should only include tags within the current scope, the mapping is only
        included if the page is a child of the page the listing was found on.

        Arguments:
            mapping: The mapping.

        Yields:
            The current tag.
        """
        assert isinstance(mapping, Mapping)

        # If the mapping is on the same page as the listing, we skip it, as
        # it makes no sense to link to a listing on the same page
        if mapping.item == self.page:
            return iter([])

        # If the listing should only include tags within the current scope, we
        # check if the page is a child of the page the listing is embedded in
        if self.config.scope:
            assert isinstance(mapping.item, Page)
            # Note that we must use `File.src_uri` here, or we won't be able to
            # detect the difference between pages, and index pages in sections,
            # as `foo/index.md` and `foo.md` look the same when using `Page.url`
            base = posixpath.dirname(posixpath.normpath(self.page.file.src_uri))
            if not mapping.item.file.src_uri.startswith(posixpath.join(base, "")):
                return iter([])

        # If an exclusion list is given, expand each tag to check if the tag
        # itself or one of its parents is excluded from the listing
        if self.config.exclude:
            if any(mapping & self.config.exclude):
                return iter([])

        # If an inclusion list is given, expand each tag to check if the tag
        # itself or one of its parents is included in the listing
        if self.config.include:
            return mapping & self.config.include

        # Otherwise, we can just return an iterator over the set of tags of the
        # mapping as is, as no expansion is required
        return iter(mapping.tags)

    # -------------------------------------------------------------------------

    page: Page
    """
    The page the listing is embedded in.
    """

    id: str
    """
    The listing identifier.

    As authors may place an arbitrary number of listings on any page, we need
    to be able to distinguish between them. This is done automatically.
    """

    config: ListingConfig
    """
    The listing configuration.
    """

    tags: dict[Tag, ListingTree]
    """
    The listing trees, each of which associated with a tag.
    """

    # -------------------------------------------------------------------------

    def add(self, mapping: Mapping, *, hidden = True) -> None:
        """
        Add mapping to listing.

        Mappings are only added to listings, if the listing features tags that
        are also featured in the mapping. The caller can decide whether hidden
        tags should be rendered or not, e.g., automatically set by the plugin
        when shadow tags are disabled.

        Arguments:
            mapping: The mapping.
            hidden: Whether to add hidden tags.
        """
        for leaf in self & mapping:
            tree = self.tags

            # Skip if hidden tags should not be rendered
            if not hidden and leaf.hidden:
                continue

            # Iterate over expanded tags
            for tag in reversed([*leaf]):
                if tag not in tree:
                    tree[tag] = ListingTree(tag)

                # If the tag is the leaf, i.e., the actual tag we want to add,
                # we add the mapping to the listing tree's mappings
                if tag == leaf:
                    tree[tag].mappings.append(mapping)

                # Otherwise, we continue walking down the tree
                else:
                    tree = tree[tag].children



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from collections.abc import Callable
from material.utilities.filter import FilterConfig
from mkdocs.config.config_options import (
    DictOfItems, Deprecated, ListOfItems, SubConfig, Type
)
from mkdocs.config.base import Config
from pymdownx.slugs import slugify

from . import item_title, tag_name
from .structure.listing import ListingConfig
from .structure.tag.options import TagSet

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Tags plugin configuration
class TagsConfig(Config):
    enabled = Type(bool, default = True)

    # Settings for filtering
    filters = SubConfig(FilterConfig)

    # Settings for tags
    tags = Type(bool, default = True)
    tags_slugify = Type(Callable, default = slugify(case = "lower"))
    tags_slugify_separator = Type(str, default = "-")
    tags_slugify_format = Type(str, default = "tag:{slug}")
    tags_hierarchy = Type(bool, default = False)
    tags_hierarchy_separator = Type(str, default = "/")
    tags_sort_by = Type(Callable, default = tag_name)
    tags_sort_reverse = Type(bool, default = False)
    tags_name_property = Type(str, default = "tags")
    tags_name_variable = Type(str, default = "tags")
    tags_allowed = TagSet()

    # Settings for listings
    listings = Type(bool, default = True)
    listings_map = DictOfItems(SubConfig(ListingConfig), default = {})
    listings_sort_by = Type(Callable, default = item_title)
    listings_sort_reverse = Type(bool, default = False)
    listings_tags_sort_by = Type(Callable, default = tag_name)
    listings_tags_sort_reverse = Type(bool, default = False)
    listings_directive = Type(str, default = "material/tags")
    listings_layout = Type(str, default = "default")
    listings_toc = Type(bool, default = True)

    # Settings for shadow tags
    shadow = Type(bool, default = False)
    shadow_on_serve = Type(bool, default = True)
    shadow_tags = TagSet()
    shadow_tags_prefix = Type(str, default = "")
    shadow_tags_suffix = Type(str, default = "")

    # Settings for export
    export = Type(bool, default = True)
    export_file = Type(str, default = "tags.json")
    export_only = Type(bool, default = False)

    # Deprecated settings
    tags_compare = Deprecated(moved_to = "tags_sort_by")
    tags_compare_reverse = Deprecated(moved_to = "tags_sort_reverse")
    tags_pages_compare = Deprecated(moved_to = "listings_sort_by")
    tags_pages_compare_reverse = Deprecated(moved_to = "listings_sort_reverse")
    tags_file = Deprecated(option_type = Type(str))
    tags_extra_files = Deprecated(
        option_type = DictOfItems(ListOfItems(Type(str)), default = {})
    )



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

import functools
import json
import logging
import os
import subprocess
import sys

from fnmatch import fnmatch
from colorama import Fore, Style
from concurrent.futures import Future
from concurrent.futures.thread import ThreadPoolExecutor
from hashlib import sha1
from mkdocs import utils
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.exceptions import PluginError
from mkdocs.plugins import BasePlugin
from mkdocs.structure.files import File
from shutil import which
try:
    from PIL import Image
except ImportError:
    pass

from .config import OptimizeConfig

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Optimize plugin
class OptimizePlugin(BasePlugin[OptimizeConfig]):
    supports_multiple_instances = True

    # Manifest
    manifest: dict[str, str] = {}

    # Initialize plugin
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Initialize incremental builds
        self.is_serve = False

    # Determine whether we're serving the site
    def on_startup(self, *, command, dirty):
        self.is_serve = command == "serve"

        # Initialize thread pool
        self.pool = ThreadPoolExecutor(self.config.concurrency)
        self.pool_jobs: dict[str, Future] = {}

    # Resolve and load manifest
    def on_config(self, config):
        if not self.config.enabled:
            return

        # Resolve cache directory (once) - this is necessary, so the cache is
        # always relative to the configuration file, and thus project, and not
        # relative to the current working directory, or it would not work with
        # the projects plugin.
        path = os.path.abspath(self.config.cache_dir)
        if path != self.config.cache_dir:
            self.config.cache_dir = os.path.join(
                os.path.dirname(config.config_file_path),
                os.path.normpath(self.config.cache_dir)
            )

            # Ensure cache directory exists
            os.makedirs(self.config.cache_dir, exist_ok = True)

        # Initialize manifest
        self.manifest_file = os.path.join(
            self.config.cache_dir, "manifest.json"
        )

        # Load manifest if it exists and the cache should be used
        if os.path.isfile(self.manifest_file) and self.config.cache:
            try:
                with open(self.manifest_file) as f:
                    self.manifest = json.load(f)
            except:
                pass

    # Initialize optimization pipeline
    def on_env(self, env, *, config, files):
        if not self.config.enabled:
            return

        # Skip if media files should not be optimized
        if not self.config.optimize:
            return

        # Filter all optimizable media files and steal reponsibility from MkDocs
        # by removing them from the files collection. Then, start a concurrent
        # job that checks if an image was already optimized and can be returned
        # from the cache, or optimize it accordingly.
        for file in files.media_files():
            if self._is_excluded(file):
                continue

            # Spawn concurrent job to optimize the given image and add future
            # to job dictionary, as it returns the file we need to copy later
            path = os.path.join(self.config.cache_dir, file.src_path)
            self.pool_jobs[file.abs_src_path] = self.pool.submit(
                self._optimize_image, file, path, config
            )

            # Steal responsibility from MkDocs
            files.remove(file)

    # Finish optimization pipeline
    def on_post_build(self, *, config):
        if not self.config.enabled:
            return

        # Skip if media files should not be optimized
        if not self.config.optimize:
            return

        # Reconcile concurrent jobs - we need to wait for all jobs to finish
        # before we can copy the optimized files to the output directory. If an
        # exception occurred in one of the jobs, we raise it here, so the build
        # fails and the author can fix the issue.
        for path, future in self.pool_jobs.items():
            if future.exception():
                raise future.exception()
            else:
                file: File = future.result()
                file.copy_file()

        # Save manifest if cache should be used
        if self.config.cache:
            with open(self.manifest_file, "w") as f:
                f.write(json.dumps(self.manifest, indent = 2, sort_keys = True))

        # Compute and print gains through optimization
        if self.config.print_gain_summary:
            print(Style.NORMAL)
            print(f"  Optimizations:")

            # Print summary for file extension
            for seek in [".png", ".jpg"]:
                size = size_opt = 0
                for path, future in self.pool_jobs.items():
                    file: File = future.result()

                    # Skip files that are not of the given type
                    _, extension = os.path.splitext(path)
                    extension = ".jpg" if extension == ".jpeg" else extension
                    if extension != seek:
                        continue

                    # Compute size before and after optimization
                    size     += os.path.getsize(path)
                    size_opt += os.path.getsize(file.abs_dest_path)

                # Compute absolute and relative gain
                if size and size_opt:
                    gain_abs = size - size_opt
                    gain_rel = (1 - size_opt / size) * 100

                    # Print summary for files
                    print(
                        f"    *{seek} {Fore.GREEN}{_size(size_opt)}"
                        f"{Fore.WHITE}{Style.DIM} ↓ "
                        f"{_size(gain_abs)} [{gain_rel:3.1f}%]"
                        f"{Style.RESET_ALL}"
                    )

            # Reset all styles
            print(Style.RESET_ALL)

    # Save manifest on shutdown
    def on_shutdown(self):
        if not self.config.enabled:
            return

        # Shutdown thread pool - if we're on Python 3.9 and above, cancel all
        # pending futures that have not yet been scheduled
        if sys.version_info >= (3, 9):
            self.pool.shutdown(cancel_futures = True)
        else:
            self.pool.shutdown()

        # Save manifest if cache should be used
        if self.manifest and self.config.cache:
            with open(self.manifest_file, "w") as f:
                f.write(json.dumps(self.manifest, indent = 2, sort_keys = True))

    # -------------------------------------------------------------------------

    # Check if a file can be optimized
    def _is_optimizable(self, file: File):

        # Check if PNG images should be optimized
        if file.url.endswith((".png")):
            return self.config.optimize_png

        # Check if JPG images should be optimized
        if file.url.endswith((".jpg", ".jpeg")):
            return self.config.optimize_jpg

        # File can not be optimized by the plugin
        return False

    # Check if the given file is excluded
    def _is_excluded(self, file: File):
        if not self._is_optimizable(file):
            return True

        # Check if file matches one of the inclusion patterns
        path = file.src_path
        if self.config.optimize_include:
            for pattern in self.config.optimize_include:
                if fnmatch(file.src_uri, pattern):
                    return False

            # File is not included
            log.debug(f"Excluding file '{path}' due to inclusion patterns")
            return True

        # Check if file matches one of the exclusion patterns
        for pattern in self.config.optimize_exclude:
            if fnmatch(file.src_uri, pattern):
                log.debug(f"Excluding file '{path}' due to exclusion patterns")
                return True

        # File is not excluded
        return False

    # Optimize image and write to cache
    def _optimize_image(self, file: File, path: str, config: MkDocsConfig):
        with open(file.abs_src_path, "rb") as f:
            data = f.read()
            hash = sha1(data).hexdigest()

            # Check if file hash changed, so we need to optimize again
            prev = self.manifest.get(file.url, "")
            if hash != prev or not os.path.isfile(path):
                os.makedirs(os.path.dirname(path), exist_ok = True)

                # Optimize PNG image using pngquant
                if file.url.endswith((".png")):
                    self._optimize_image_png(file, path, config)

                # Optimize JPG image using pillow
                if file.url.endswith((".jpg", ".jpeg")):
                    self._optimize_image_jpg(file, path, config)

                # Compute size before and after optimization
                size     = len(data)
                size_opt = os.path.getsize(path)

                # Compute absolute and relative gain
                gain_abs = size - size_opt
                gain_rel = (1 - size_opt / size) * 100

                # Print how much we gained, if we did and desired
                gain = ""
                if gain_abs and self.config.print_gain:
                    gain += " ↓ "
                    gain += " ".join([_size(gain_abs), f"[{gain_rel:3.1f}%]"])

                # Print summary for file
                log.info(
                    f"Optimized media file: {file.src_uri} "
                    f"{Fore.GREEN}{_size(size_opt)}"
                    f"{Fore.WHITE}{Style.DIM}{gain}"
                    f"{Style.RESET_ALL}"
                )

                # Update manifest by associating file with hash
                self.manifest[file.url] = hash

        # Compute project root
        root = os.path.dirname(config.config_file_path)

        # Compute source file system path
        file.abs_src_path = path
        file.src_path     = os.path.relpath(path, root)

        # Return file to be copied from cache
        return file

    # Optimize PNG image - we first tried to use libimagequant, but encountered
    # the occassional segmentation fault, which means it's probably not a good
    # choice. Instead, we just rely on pngquant which seems much more stable.
    def _optimize_image_png(self, file: File, path: str, config: MkDocsConfig):

        # Check if the required dependencies for optimizing are available, which
        # is, at the absolute minimum, the 'pngquant' binary, and raise an error
        # to the caller, so he can decide what to do with the error. The caller
        # can treat this as a warning or an error to abort the build.
        if not which("pngquant"):
            docs = os.path.relpath(config.docs_dir)
            path = os.path.relpath(file.abs_src_path, docs)
            raise PluginError(
                f"Couldn't optimize image '{path}' in '{docs}': 'pngquant' "
                f"not found. Make sure 'pngquant' is installed and in your path"
            )

        # Build command line arguments
        args = ["pngquant",
            "--force", "--skip-if-larger",
            "--output", path,
            "--speed", f"{self.config.optimize_png_speed}"
        ]

        # Add flag to remove optional metadata
        if self.config.optimize_png_strip:
            args.append("--strip")

        # Set input file and run, then check if pngquant actually wrote a file,
        # as we instruct it not to if the size of the optimized file is larger.
        # This can happen if files are already compressed and optimized by
        # the author. In that case, just copy the original file.
        subprocess.run([*args, file.abs_src_path])
        if not os.path.isfile(path):
            utils.copy_file(file.abs_src_path, path)

    # Optimize JPG image
    def _optimize_image_jpg(self, file: File, path: str, config: MkDocsConfig):

        # Check if the required dependencies for optimizing are available, which
        # is, at the absolute minimum, the 'pillow' package, and raise an error
        # to the caller, so he can decide what to do with the error. The caller
        # can treat this as a warning or an error to abort the build.
        if not _supports("Image"):
            docs = os.path.relpath(config.docs_dir)
            path = os.path.relpath(file.abs_src_path, docs)
            raise PluginError(
                f"Couldn't optimize image '{path}' in '{docs}': install "
                f"required dependencies – pip install 'mkdocs-material[imaging]'"
            )

        # Open and save optimized image
        image = Image.open(file.abs_src_path)
        image.save(path, "jpeg",
            quality     = self.config.optimize_jpg_quality,
            progressive = self.config.optimize_jpg_progressive
        )

# -----------------------------------------------------------------------------
# Helper functions
# -----------------------------------------------------------------------------

# Check for presence of optional imports
@functools.lru_cache(maxsize = None)
def _supports(name: str):
    return name in globals()

# -----------------------------------------------------------------------------

# Print human-readable size
def _size(value):
    for unit in ["B", "kB", "MB", "GB", "TB", "PB", "EB", "ZB"]:
        if abs(value) < 1000.0:
            return f"{value:3.1f} {unit}"
        value /= 1000.0

# -----------------------------------------------------------------------------
# Data
# -----------------------------------------------------------------------------

# Set up logging
log = logging.getLogger("mkdocs.material.optimize")



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

import re

from mkdocs.config.base import Config
from mkdocs.config.config_options import (
    Choice, DictOfItems, ListOfItems, SubConfig, Type
)
try:
    from PIL.Image import Image as _Image
except ImportError:
    pass

# -----------------------------------------------------------------------------
# Options
# -----------------------------------------------------------------------------

# Options for origin
Origin = (
    "start top",        "center top",       "end top",
    "start center",     "center",           "end center",
    "start bottom",     "center bottom",    "end bottom",
    "start",                                "end"
)

# Options for overflow
Overflow = (
    "truncate",
    "shrink"
)

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Size
class Size(Config):
    width = Type(int, default = 0)
    height = Type(int, default = 0)

# Offset
class Offset(Config):
    x = Type(int, default = 0)
    y = Type(int, default = 0)

# # -----------------------------------------------------------------------------

# Background
class Background(Config):
    color = Type(str, default = "")
    image = Type(str, default = "")

# # -----------------------------------------------------------------------------

# Icon
class Icon(Config):
    value = Type(str, default = "")
    color = Type(str, default = "")

# # -----------------------------------------------------------------------------

# Line
class Line(Config):
    amount = Type((int, float), default = 1)
    height = Type((int, float), default = 1)

# Font
class Font(Config):
    family = Type(str, default = "Roboto")
    variant = Type(str, default = "")
    style = Type(str, default = "Regular")

# Typography
class Typography(Config):
    content = Type(str, default = "")
    align = Choice(Origin, default = "start top")
    overflow = Choice(Overflow, default = "truncate")
    color = Type(str, default = "")
    line = SubConfig(Line)
    font = SubConfig(Font)

# -----------------------------------------------------------------------------

# Layer
class Layer(Config):
    size = SubConfig(Size)
    offset = SubConfig(Offset)
    origin = Choice(Origin, default = "start top")
    background = SubConfig(Background)
    icon = SubConfig(Icon)
    typography = SubConfig(Typography)

# -----------------------------------------------------------------------------

# Layout
class Layout(Config):
    definitions = ListOfItems(Type(str), default = [])
    tags = DictOfItems(Type(str), default = {})
    size = SubConfig(Size)
    layers = ListOfItems(SubConfig(Layer), default = [])

# -----------------------------------------------------------------------------
# Functions
# -----------------------------------------------------------------------------

# Get layer or layout size as tuple
def get_size(layer: Layer | Layout):
    return layer.size.width, layer.size.height

# Get layer offset as tuple
def get_offset(layer: Layer, image: _Image):
    x, y = layer.offset.x, layer.offset.y

    # Compute offset from origin - if an origin is given, compute the offset
    # relative to the image and layer size to allow for flexible positioning
    if layer.origin != "start top":
        origin = re.split(r"\s+", layer.origin)

        # Get layer size
        w, h = get_size(layer)

        # Compute origin on x-axis
        if   "start"  in origin: pass
        elif "end"    in origin: x += (image.width  - w)  - 2 * x
        elif "center" in origin: x += (image.width  - w) >> 1

        # Compute origin on y-axis
        if   "top"    in origin: pass
        elif "bottom" in origin: y += (image.height - h)  - 2 * y
        elif "center" in origin: y += (image.height - h) >> 1

    # Return offset
    return x, y



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os

from mkdocs.config.base import Config
from mkdocs.config.config_options import Deprecated, ListOfItems, Type
from mkdocs.config.defaults import _LogLevel

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Social plugin configuration
class SocialConfig(Config):
    enabled = Type(bool, default = True)
    concurrency = Type(int, default = max(1, os.cpu_count() - 1))

    # Settings for caching
    cache = Type(bool, default = True)
    cache_dir = Type(str, default = ".cache/plugin/social")

    # Settings for logging
    log = Type(bool, default = True)
    log_level = _LogLevel(default = "warn")

    # Settings for cards
    cards = Type(bool, default = True)
    cards_dir = Type(str, default = "assets/images/social")
    cards_layout_dir = Type(str, default = "layouts")
    cards_layout = Type(str, default = "default")
    cards_layout_options = Type(dict, default = {})
    cards_include = ListOfItems(Type(str), default = [])
    cards_exclude = ListOfItems(Type(str), default = [])

    # Settings for debugging
    debug = Type(bool, default = False)
    debug_on_build = Type(bool, default = False)
    debug_grid = Type(bool, default = True)
    debug_grid_step = Type(int, default = 32)
    debug_color = Type(str, default = "grey")

    # Deprecated settings
    cards_color = Deprecated(
        message =
            "Deprecated, use 'cards_layout_options.background_color' "
            "and 'cards_layout_options.color' with 'default' layout"
        )
    cards_font = Deprecated(
        message = "Deprecated, use 'cards_layout_options.font_family'"
    )



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

import logging
import multiprocessing
import os
import posixpath

from collections.abc import Callable
from concurrent.futures import Future, as_completed
from concurrent.futures.process import ProcessPoolExecutor
from logging import Logger
from material.plugins.projects.config import ProjectsConfig
from material.plugins.projects.structure import Project, ProjectJob
from mkdocs.commands.build import build
from mkdocs.config.base import ConfigErrors, ConfigWarnings
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.exceptions import Abort
from mkdocs.livereload import LiveReloadServer
from urllib.parse import urlparse

from .log import (
    get_log_for,
    get_log_formatter,
    get_log_handler,
    get_log_level_for
)
from .watcher import ProjectsWatcher

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Projects builder
class ProjectsBuilder:

    # Set of projects
    projects: set[Project] = set()

    # Projects watcher
    watcher: ProjectsWatcher | None = None

    # Initialize projects builder - compared to our other concurrent plugins,
    # this plugin is forced to use a process pool in order to guarantee proper
    # isolation, as MkDocs itself is not thread-safe. We also need to recreate
    # the process pool on every build, or CTRL-C is broken
    def __init__(self, config: MkDocsConfig, plugin: ProjectsConfig):

        # Initialize root project
        self.root = Project(config.config_file_path, plugin)
        self.root.config = config

        # Initialize process pool
        self.pool = ProcessPoolExecutor
        self.pool_jobs: dict[str, Future] = {}

    # Build projects
    def build(self, serve: bool = False, dirty: bool = False):
        self.pool = ProcessPoolExecutor(
            self.root.plugin.concurrency,
            mp_context = multiprocessing.get_context("spawn")
        )

        # Determine projects in topological order and prepare for building
        built: list[str] = []
        queue = [*self.root.jobs()]
        for job in queue:
            _setup(job.project, self.root, serve)
            if serve:
                self._link(job.project)

        # Schedule projects for building
        for job in reversed(queue):
            if self._schedule(job, serve, dirty):
                queue.remove(job)

        # Build loop - iteratively build more projects while there are still
        # projects to be built, sticking to the topological order.
        while len(built) < len(self.pool_jobs):
            for future in as_completed(self.pool_jobs.values()):
                slug, errors, warnings = future.result()
                if slug in built:
                    continue

                # Mark project as built
                built.append(slug)

                # Schedule projects for building
                for job in reversed(queue):
                    if self._schedule(job, serve, dirty):
                        queue.remove(job)

                # Print errors and warnings
                for project in self.projects:
                    if project.slug == slug:
                        _print(get_log_for(project), errors, warnings)
                        break

        # Shutdown process pool
        self.pool.shutdown()
        if self.watcher:

            # Update watched paths
            for project in self.projects:
                if project.slug not in built:
                    self.pool_jobs.pop(project.slug, None)
                    self.watcher.unwatch(project)
                else:
                    self.watcher.watch(project, self.taint)

    # Taint a project to schedule it for building
    def taint(self, project: Project):
        self.pool_jobs.pop(project.slug, None)

    # Watch and serve projects
    def serve(self, server: LiveReloadServer, is_dirty: bool = False):
        self.watcher = ProjectsWatcher(server)
        self.watcher.watch(self.root, self.taint)
        for project in self.projects:
            self.watcher.watch(project, self.taint)

    # -------------------------------------------------------------------------

    # Create symlink for project if we're serving the site
    def _link(self, project: Project):

        # Compute path for slug from current project - normalize path,
        # as paths computed from slugs or site URLs use forward slashes
        path = project.path(self.root)
        path = os.path.normpath(path)

        # Create symbolic link, if we haven't already
        path = os.path.join(self.root.config.site_dir, path)
        if not os.path.islink(path):

            # Ensure link target exists
            target = os.path.realpath(os.path.dirname(path))
            if not os.path.exists(target):
                os.makedirs(target, exist_ok = True)

            # Create symbolic link
            os.makedirs(os.path.dirname(path), exist_ok = True)
            os.symlink(project.config.site_dir, path)

    # Schedule project for building - spawn concurrent job to build the project
    # and add a future to the jobs dictionary to link build results to projects
    def _schedule(self, job: ProjectJob, serve: bool, dirty: bool):
        self.projects.add(job.project)

        # Exit early, if project doesn't need to be re built
        if job.project.slug in self.pool_jobs:
            return True

        # Check if dependencies have been built already, and if so, remove
        # them from the list of dependencies. If a dependency has failed to
        # build, we'll raise an exception, which will be caught by the main
        # process, and the entire build will be aborted.
        for dependency in [*job.dependencies]:
            future = self.pool_jobs[dependency.slug]
            if future.running():
                continue

            # If the dependency has failed to build, we'll raise an exception
            # to abort the entire build, as we can't build the project itself
            # without the dependency. This will be caught by the main process.
            # Otherwise, we'll remove the dependency from the list.
            if future.exception():
                raise future.exception()
            elif future.done():
                job.dependencies.remove(dependency)

        # If all dependencies of the project have been built, we can build
        # the project itself by spawning a concurrent job
        if not job.dependencies:
            self.pool_jobs[job.project.slug] = self.pool.submit(
                _build, job.project, serve, dirty,
                get_log_level_for(job.project)
            )

        # Return whether the project has been scheduled
        return not job.dependencies

# -----------------------------------------------------------------------------
# Helper functions
# -----------------------------------------------------------------------------

# Setup project by preparing it for building
def _setup(project: Project, root: Project, serve: bool):
    assert project.slug != "."

    # Retrieve configuration of top-level project and transform project
    transform = root.plugin.projects_config_transform
    if isinstance(transform, Callable):
        transform(project, root)

    # If the top-level project defines a site URL, we need to make sure that the
    # site URL of the project is set as well, setting it to the path we derive
    # from the slug. This allows to define the URL independent of the entire
    # project's directory structure. If the top-level project doesn't define a
    # site URL, it might be the case that the author is building a consolidated
    # project of several nested projects that are independent, but which should
    # be bundled together for distribution. As this is a case that is quite
    # common, we're not raising a warning or error.
    path = project.path(root)
    if root.config.site_url:

        # If the project doesn't have a site URL, compute it from the site URL
        # of the top-level project and the path derived from the slug
        if not project.config.site_url:
            project.config.site_url = posixpath.join(
                root.config.site_url,
                path
            )

        # If we're serving the site, replace the project's host name with the
        # dev server address, so we can serve nested projects as well
        if serve:
            url = urlparse(project.config.site_url)
            url = url._replace(
                scheme = "http",
                netloc = str(root.config.dev_addr)
            )

            # Update site URL with dev server address
            project.config.site_url = url.geturl()

    # If we're building the site, the project's output must be written to the
    # site directory of the top-level project, so we can serve it from there
    if not serve:
        project.config.site_dir = os.path.join(
            root.config.site_dir,
            os.path.normpath(path)
        )

    # If we're serving the site, we must fall back to symbolic links, as MkDocs
    # will empty the entire site directory every time it performs a build
    else:
        project.config.site_dir = os.path.join(
            os.path.dirname(project.config.config_file_path),
            project.config.site_dir
        )

# Build project - note that regardless of whether MkDocs was started in build
# or serve mode, projects must always be built, as they're served by the root
def _build(project: Project, serve: bool, dirty: bool, level = logging.WARN):
    config = project.config

    # Change working directory to project root - this is necessary, or relative
    # paths used in extensions and plugins will be resolved incorrectly
    os.chdir(os.path.dirname(config.config_file_path))

    # Validate configuration
    errors, warnings = config.validate()
    if not errors:

        # Retrieve and configure MkDocs' logger
        log = logging.getLogger("mkdocs")
        log.setLevel(level)

        # Hack: there seems to be an inconsistency between operating systems,
        # and it's yet unclear where this is coming from - on macOS, the MkDocs
        # default logger has no designated handler registered, but on Linux it
        # does. If there's no handler, we need to create one. If there is, we
        # must only set the formatter, as otherwise we'll end up with the same
        # message printed on two log handlers - see https://t.ly/q7UEq
        handler = get_log_handler(project)
        if not log.hasHandlers():
            log.addHandler(handler)
        else:
            for handler in log.handlers:
                handler.setFormatter(get_log_formatter(project))

        # Build project and dispatch startup and shutdown plugin events - note
        # that we must pass the correct command to the event handler, but run
        # the build command anyway, because otherwise some plugins will not
        # run in serve mode.
        command = "serve" if serve else "build"
        config.plugins.run_event("startup", command = command, dirty = dirty)
        try:
            build(config, dirty = dirty)
        finally:
            config.plugins.run_event("shutdown")
            log.removeHandler(handler)

    # Return slug, errors and warnings
    return project.slug, errors, warnings

# Print errors and warnings resulting from building a project
def _print(log: Logger, errors: ConfigErrors, warnings: ConfigWarnings):

    # Print warnings
    for value, message in warnings:
        log.warning(f"Config value '{value}': {message}")

    # Print errors
    for value, message in errors:
        log.error(f"Config value '{value}': {message}")

    # Abort if there were errors
    if errors:
        raise Abort(f"Aborted with {len(errors)} configuration errors")



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

from collections.abc import Iterator
from functools import total_ordering

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

@total_ordering
class Tag:
    """
    A tag.

    Tags can be used to categorize pages and group them into a tag structure. A
    tag is a simple string, which can be split into a hierarchy of tags by using
    the character or string as defined in the `hierarchy_separator` setting in
    `mkdocs.yml`. Each parent tag contains their child tags.

    Example:

    ```yaml
    tags:
      - foo/bar
      - foo/baz
      - qux
    ```

    The tag structure for the above example would look like this:

    ```
    .
    ├─ foo
    │  ├─ bar
    │  └─ baz
    └─ qux
    ```

    Note that this class does not split the tag name into a hierarchy of tags
    by itself, but rather provides a simple interface to iterate over the tag
    and its parents. Splitting is left to the caller, in order to allow for
    changing the separator in `mkdocs.yml`.
    """

    def __init__(
        self, name: str, *, parent: Tag | None = None, hidden = False
    ):
        """
        Initialize the tag.

        Arguments:
            name: The tag name.
            parent: The parent tag.
            hidden: Whether the tag is hidden.
        """
        self.name = name
        self.parent = parent
        self.hidden = hidden

    def __repr__(self) -> str:
        """
        Return a printable representation of the tag.

        Returns:
            Printable representation.
        """
        return f"Tag('{self.name}')"

    def __str__(self) -> str:
        """
        Return a string representation of the tag.

        Returns:
            String representation.
        """
        return self.name

    def __hash__(self) -> int:
        """
        Return the hash of the tag.

        Returns:
            The hash.
        """
        return hash(self.name)

    def __iter__(self) -> Iterator[Tag]:
        """
        Iterate over the tag and its parent tags.

        Note that the first tag returned is the tag itself, followed by its
        parent tags in ascending order. This allows to iterate over the tag
        and its parents in a single loop, which is useful for generating
        tree or breadcrumb structures.

        Yields:
            The current tag.
        """
        tag = self
        while tag:
            yield tag
            tag = tag.parent

    def __contains__(self, other: Tag) -> bool:
        """
        Check if the tag contains another tag.

        Arguments:
            other: The other tag to check.

        Returns:
            Whether the tag contains the other tag.
        """
        assert isinstance(other, Tag)
        return any(tag == other for tag in self)

    def __eq__(self, other: Tag) -> bool:
        """
        Check if the tag is equal to another tag.

        Arguments:
            other: The other tag to check.

        Returns:
            Whether the tags are equal.
        """
        assert isinstance(other, Tag)
        return self.name == other.name

    def __lt__(self, other: Tag) -> bool:
        """
        Check if the tag is less than another tag.

        Arguments:
            other: The other tag to check.

        Returns:
            Whether the tag is less than the other tag.
        """
        assert isinstance(other, Tag)
        return self.name < other.name

    # -------------------------------------------------------------------------

    name: str
    """
    The tag name.
    """

    parent: Tag | None
    """
    The parent tag.
    """

    hidden: bool
    """
    Whether the tag is hidden.
    """



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

from collections.abc import Iterator
from material.plugins.tags.config import TagsConfig
from material.plugins.tags.structure.mapping import Mapping
from material.plugins.tags.structure.tag import Tag
from material.plugins.tags.structure.tag.options import TagSet
from mkdocs.structure.pages import Page

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

class MappingManager:
    """
    A mapping manager.

    The mapping manager is responsible for collecting all tags from the front
    matter of pages, and for building a tag structure from them, nothing more.
    """

    def __init__(self, config: TagsConfig):
        """
        Initialize the mapping manager.

        Arguments:
            config: The configuration.
        """
        self.config = config
        self.format = TagSet(allowed = self.config.tags_allowed)
        self.data = {}

    def __repr__(self) -> str:
        """
        Return a printable representation of the mapping manager.

        Returns:
            Printable representation.
        """
        return _print(self)

    def __iter__(self) -> Iterator[Mapping]:
        """
        Iterate over mappings.

        Yields:
            The current mapping.
        """
        return iter(self.data.values())

    # -------------------------------------------------------------------------

    config: TagsConfig
    """
    The configuration.
    """

    format: TagSet
    """
    The mapping format.

    This is the validator that is used to check if tags are valid, including
    the tags in the front matter of pages, as well as the tags defined in the
    configuration. Numbers and booleans are always converted to strings before
    creating tags, and the allow list is checked as well, if given.
    """

    data: dict[str, Mapping]
    """
    The mappings.
    """

    # -------------------------------------------------------------------------

    def add(self, page: Page, markdown: str) -> Mapping | None:
        """
        Add page.

        This method is called by the tags plugin to retrieve all tags of a page.
        It extracts all tags from the front matter of the given page, and adds
        them to the mapping. If no tags are found, no mapping is created and
        nothing is returned.

        Note that this method is intended to be called with the page during the
        `on_page_markdown` event, as it reads the front matter of a page. Also,
        the Markdown must be explicitly passed, as we could otherwise run into
        inconsistencies when other plugins modify the Markdown.

        Arguments:
            page: The page.
            markdown: The page's Markdown.

        Returns:
            The mapping or nothing.
        """
        assert isinstance(markdown, str)

        # Return nothing if page doesn't have tags
        tags = self.config.tags_name_property
        if not page.meta.get(tags, []):
            return

        # Create mapping and associate with page
        mapping = Mapping(page)
        self.data[page.url] = mapping

        # Retrieve and validate tags, and add to mapping
        for tag in self.format.validate(page.meta[tags]):
            mapping.tags.add(self._configure(tag))

        # Return mapping
        return mapping

    def get(self, page: Page) -> Mapping | None:
        """
        Get mapping for page, if any.

        Arguments:
            page: The page.

        Returns:
            The mapping or nothing.
        """
        if page.url in self.data:
            return self.data[page.url]

    # -------------------------------------------------------------------------

    def _configure(self, tag: Tag) -> Tag:
        """
        Configure tag.

        This method is called by the mapping manager to configure a tag for the
        the tag structure. Depending on the configuration, the tag is expanded
        into a hierarchy of tags, and can be marked as hidden if it is a shadow
        tag, hiding it from mappings and listings when rendering.

        Arguments:
            tag: The tag.

        Returns:
            The configured tag.
        """
        if self.config.tags_hierarchy:
            return self._configure_hierarchy(tag)
        else:
            return self._configure_shadow(tag, tag.name)

    def _configure_hierarchy(self, tag: Tag) -> Tag:
        """
        Configure hierarchical tag.

        Note that shadow tags that occur as part of a tag hierarchy propagate
        their hidden state to all of their children.

        Arguments:
            tag: The tag.

        Returns:
            The configured tag.
        """
        separator = self.config.tags_hierarchy_separator
        root, *rest = tag.name.split(separator)

        # Create tag root and hierarchy
        tag = self._configure_shadow(Tag(root), root)
        for name in rest:
            tag = self._configure_shadow(Tag(
                separator.join([tag.name, name]),
                parent = tag, hidden = tag.hidden
            ), name)

        # Return tag
        return tag

    def _configure_shadow(self, tag: Tag, name: str) -> Tag:
        """
        Configure shadow tag.

        Regardless of the configuration, tags are always marked as hidden if
        they're classified as shadow tags, e.g., if their name matches the
        configured shadow prefix or suffix, or if they're part of the list of
        shadow tags. Whether they're displayed is decided before rendering.

        The tag name must be passed separately, as it may be different from the
        tag's name, e.g., when creating a tag hierarchy. In this case, the name
        represents the part that was added to the tag, essentially the suffix.
        The name is checked for shadow prefixes and suffixes.

        Arguments:
            tag: The tag.
            name: The tag name.

        Returns:
            The configured tag.
        """
        if not tag.hidden:
            tag.hidden = tag in self.config.shadow_tags

        # Check if tag matches shadow prefix, if defined
        if not tag.hidden and self.config.shadow_tags_prefix:
            tag.hidden = name.startswith(self.config.shadow_tags_prefix)

        # Check if tag matches shadow suffix, if defined
        if not tag.hidden and self.config.shadow_tags_suffix:
            tag.hidden = name.endswith(self.config.shadow_tags_suffix)

        # Return tag
        return tag

# -----------------------------------------------------------------------------
# Functions
# -----------------------------------------------------------------------------

def _print(manager: MappingManager, indent: int = 0) -> str:
    """
    Return a printable representation of a mapping manager.

    Arguments:
        manager: The mapping manager.
        indent: The indentation level.

    Returns:
        Printable representation.
    """
    lines: list[str] = []
    lines.append(" " * indent + f"MappingManager()")

    # Print mappings
    for mapping in manager:
        lines.append(" " * (indent + 2) + repr(mapping))

    # Concatenate everything
    return "\n".join(lines)



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import sys

from colorama import Fore, Style
from pathlib import Path
from sys import stderr

# Check if we're running MkDocs. Disable the warning automatically for other
# contexts, e.g., to allow for forks of MkDocs to adjust behavior to fit
def is_mkdocs():
    path = Path(sys.argv[0])
    return path.name in ("mkdocs", "mkdocs.exe") or (
        path.name == "__main__.py" and path.parent.name == "mkdocs"
    )

# Check if we're running MkDocs and the warning hasn't been disabled
if is_mkdocs() and not os.getenv("NO_MKDOCS_2_WARNING"):
    print(
        "\n"
        f"{Fore.RED} │  ⚠  Warning from the Material for MkDocs team{Style.RESET_ALL}\n"
        f"{Fore.RED} │{Style.RESET_ALL}\n"
        f"{Fore.RED} │{Style.RESET_ALL}  MkDocs 2.0, the underlying framework of Material for MkDocs,\n"
        f"{Fore.RED} │{Style.RESET_ALL}  will introduce backward-incompatible changes, including:\n"
        f"{Fore.RED} │{Style.RESET_ALL}\n"
        f"{Fore.RED} │  × {Style.RESET_ALL}All plugins will stop working – the plugin system has been removed\n"
        f"{Fore.RED} │  × {Style.RESET_ALL}All theme overrides will break – the theming system has been rewritten\n"
        f"{Fore.RED} │  × {Style.RESET_ALL}No migration path exists – existing projects cannot be upgraded\n"
        f"{Fore.RED} │  × {Style.RESET_ALL}Closed contribution model – community members can't report bugs\n"
        f"{Fore.RED} │  × {Style.RESET_ALL}Currently unlicensed – unsuitable for production use\n"
        f"{Fore.RED} │{Style.RESET_ALL}\n"
        f"{Fore.RED} │{Style.RESET_ALL}  Our full analysis:\n"
        f"{Fore.RED} │{Style.RESET_ALL}\n"
        f"{Fore.RED} │{Style.RESET_ALL}  \033[4mhttps://squidfunk.github.io/mkdocs-material/blog/2026/02/18/mkdocs-2.0/{Style.RESET_ALL}\n"
        f"{Style.RESET_ALL}",
        file=stderr
    )

# Disable for subsequent imports
os.environ["NO_MKDOCS_2_WARNING"] = "true"



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

from markdown.treeprocessors import Treeprocessor
from mkdocs.structure.pages import Page
from mkdocs.utils import get_relative_url
from xml.etree.ElementTree import Element

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Excerpt tree processor
class ExcerptTreeprocessor(Treeprocessor):

    # Initialize excerpt tree processor
    def __init__(self, page: Page, base: Page | None = None):
        self.page = page
        self.base = base

    # Transform HTML after Markdown processing
    def run(self, root: Element):
        assert self.base
        main = True

        # We're only interested in anchors, which is why we continue when the
        # link does not start with an anchor tag
        for el in root.iter("a"):
            anchor = el.get("href")
            if not anchor.startswith("#"):
                continue

            # The main headline should link to the post page, not to a specific
            # anchor, which is why we remove the anchor in that case
            path = get_relative_url(self.page.url, self.base.url)
            if main:
                el.set("href", path)
            else:
                el.set("href", path + anchor)

            # Main headline has been seen
            main = False



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from collections.abc import Callable
from mkdocs.config.config_options import Choice, Deprecated, Optional, Type
from mkdocs.config.base import Config
from pymdownx.slugs import slugify

from . import view_name

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Blog plugin configuration
class BlogConfig(Config):
    enabled = Type(bool, default = True)

    # Settings for blog
    blog_dir = Type(str, default = "blog")
    blog_toc = Type(bool, default = False)

    # Settings for posts
    post_dir = Type(str, default = "{blog}/posts")
    post_date_format = Type(str, default = "long")
    post_url_date_format = Type(str, default = "yyyy/MM/dd")
    post_url_format = Type(str, default = "{date}/{slug}")
    post_url_max_categories = Type(int, default = 1)
    post_slugify = Type(Callable, default = slugify(case = "lower"))
    post_slugify_separator = Type(str, default = "-")
    post_excerpt = Choice(["optional", "required"], default = "optional")
    post_excerpt_max_authors = Type(int, default = 1)
    post_excerpt_max_categories = Type(int, default = 5)
    post_excerpt_separator = Type(str, default = "<!-- more -->")
    post_readtime = Type(bool, default = True)
    post_readtime_words_per_minute = Type(int, default = 265)

    # Settings for archive
    archive = Type(bool, default = True)
    archive_name = Type(str, default = "blog.archive")
    archive_date_format = Type(str, default = "yyyy")
    archive_url_date_format = Type(str, default = "yyyy")
    archive_url_format = Type(str, default = "archive/{date}")
    archive_pagination = Optional(Type(bool))
    archive_pagination_per_page = Optional(Type(int))
    archive_toc = Optional(Type(bool))

    # Settings for categories
    categories = Type(bool, default = True)
    categories_name = Type(str, default = "blog.categories")
    categories_url_format = Type(str, default = "category/{slug}")
    categories_slugify = Type(Callable, default = slugify(case = "lower"))
    categories_slugify_separator = Type(str, default = "-")
    categories_sort_by = Type(Callable, default = view_name)
    categories_sort_reverse = Type(bool, default = False)
    categories_allowed = Type(list, default = [])
    categories_pagination = Optional(Type(bool))
    categories_pagination_per_page = Optional(Type(int))
    categories_toc = Optional(Type(bool))

    # Settings for authors
    authors = Type(bool, default = True)
    authors_file = Type(str, default = "{blog}/.authors.yml")
    authors_profiles = Type(bool, default = False)
    authors_profiles_name = Type(str, default = "blog.authors")
    authors_profiles_url_format = Type(str, default = "author/{slug}")
    authors_profiles_pagination = Optional(Type(bool))
    authors_profiles_pagination_per_page = Optional(Type(int))
    authors_profiles_toc = Optional(Type(bool))

    # Settings for pagination
    pagination = Type(bool, default = True)
    pagination_per_page = Type(int, default = 10)
    pagination_url_format = Type(str, default = "page/{page}")
    pagination_format = Type(str, default = "~2~")
    pagination_if_single_page = Type(bool, default = False)
    pagination_keep_content = Type(bool, default = False)

    # Settings for drafts
    draft = Type(bool, default = False)
    draft_on_serve = Type(bool, default = True)
    draft_if_future_date = Type(bool, default = False)

    # Deprecated settings
    pagination_template = Deprecated(moved_to = "pagination_format")



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

from material.plugins.tags.structure.listing import Listing
from material.plugins.tags.structure.tag import Tag
from mkdocs.structure.pages import Page
from mkdocs.structure.toc import AnchorLink
from typing import Callable

# -----------------------------------------------------------------------------
# Typings
# -----------------------------------------------------------------------------

Slugify = Callable[[Tag], str]
"""
Slugify function.

Arguments:
    tag: The tag.

Returns:
    The slugified tag.
"""

# -----------------------------------------------------------------------------
# Functions
# -----------------------------------------------------------------------------

def populate(listing: Listing, slugify: Slugify) -> dict[Tag, AnchorLink]:
    """
    Populate the table of contents of the page the listing is embedded.

    Arguments:
        listing: The listing.
        slugify: Slugify function.

    Returns:
        The mapping of tags to anchor links.
    """
    anchors: dict[Tag, AnchorLink] = {}

    # Find injection point for listing
    host, at = find(listing)
    if at == -1:
        return anchors

    # Create anchor links
    for tree in listing:

        # Iterate over expanded tags
        for i, tag in enumerate(reversed([*tree.tag])):
            if tag not in anchors:
                level = host.level + 1 + i

                # Create anchor link
                anchors[tag] = AnchorLink(tag.name, slugify(tag), level)
                if not tag.parent:
                    continue

                # Relate anchor link to parent
                anchors[tag.parent].children.append(anchors[tag])

    # Filter top-level anchor links and insert them into the page
    children = [anchors[tag] for tag in anchors if not tag.parent]
    if listing.config.toc:
        host.children[at:at + 1] = children
    else:
        host.children.pop(at)

    # Return mapping of tags to anchor links
    return anchors

# -----------------------------------------------------------------------------

def find(listing: Listing) -> tuple[AnchorLink | None, int]:
    """
    Find anchor link for the given listing.

    This function traverses the table of contents of the given page and returns
    the anchor's parent and index of the anchor with the given identifier. If
    the anchor is on the root level, and the anchor we're looking for is an
    injection point, an anchor to host the tags is created and returned.

    Arguments:
        lising: The listing.

    Returns:
        The anchor and index.
    """
    page = listing.page

    # Traverse table of contents
    stack = list(page.toc)
    while stack:
        anchor = stack.pop()

        # Traverse children
        for i, child in enumerate(anchor.children):
            if child.id.startswith(listing.id):
                return anchor, i

            # Add child to stack
            stack.append(child)

    # Check if anchor is on the root level
    for i, anchor in enumerate(page.toc):
        if anchor.id.startswith(listing.id):

            # Create anchor link
            host = AnchorLink(page.title, page.url, 1)
            host.children = page.toc.items
            return host, i

    # Anchor could not be found
    return None, -1



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

import logging
import os
import posixpath
import re
import yaml

from collections.abc import Iterable, Iterator
from material.plugins.tags.config import TagsConfig
from material.plugins.tags.renderer import Renderer
from material.plugins.tags.structure.listing import Listing, ListingConfig
from material.plugins.tags.structure.listing.tree import ListingTree
from material.plugins.tags.structure.mapping import Mapping
from material.plugins.tags.structure.tag import Tag
from material.plugins.tags.structure.tag.reference import TagReference
from mkdocs.exceptions import PluginError
from mkdocs.structure.pages import Page
from mkdocs.structure.nav import Link
from re import Match
from urllib.parse import urlparse

from . import toc

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

class ListingManager:
    """
    A listing manager.

    The listing manager collects all listings from the Markdown of pages, then
    populates them with mappings, and renders them. Furthermore, the listing
    manager allows to obtain tag references for a given mapping, which are
    tags annotated with links to listings.
    """

    def __init__(self, config: TagsConfig, depth: int = 6):
        """
        Initialize the listing manager.

        Arguments:
            config: The configuration.
        """
        self.config = config
        self.data = set()
        self.depth = depth

    def __repr__(self) -> str:
        """
        Return a printable representation of the listing manager.

        Returns:
            Printable representation.
        """
        return _print(self)

    def __iter__(self) -> Iterator[Listing]:
        """
        Iterate over listings.

        Yields:
            The current listing.
        """
        return iter(self.data)

    def __and__(self, mapping: Mapping) -> Iterator[TagReference]:
        """
        Iterate over the tag references for the mapping.

        Arguments:
            mapping: The mapping.

        Yields:
            The current tag reference.
        """
        assert isinstance(mapping, Mapping)

        # Iterate over sorted tags and associate tags with listings - note that
        # we sort the listings for the mapping by closeness, so that the first
        # listing in the list is the closest one to the page or link the
        # mapping is associated with
        listings = self.closest(mapping)
        for tag in self._sort_tags(mapping.tags):
            ref = TagReference(tag)

            # Iterate over listings and add links
            for listing in listings:
                if tag in listing & mapping:
                    value = listing.page.url or "."

                    # Compute URL for link - make sure to remove fragments, as
                    # they may be present in links extracted from remote tags.
                    # Additionally, we need to fallback to `.` if the URL is
                    # empty (= homepage) or the links will be incorrect.
                    url = urlparse(value, allow_fragments = False)
                    url = url._replace(fragment = self._slugify(tag))

                    # Add listing link to tag reference
                    ref.links.append(
                        Link(listing.page.title, url.geturl())
                    )

            # Yield tag reference
            yield ref

    # -------------------------------------------------------------------------

    config: TagsConfig
    """
    The configuration.
    """

    data: set[Listing]
    """
    The listings.
    """

    depth: int
    """
    Table of contents maximum depth.
    """

    # -------------------------------------------------------------------------

    def add(self, page: Page, markdown: str) -> str:
        """
        Add page.

        This method is called by the tags plugin to retrieve all listings of a
        page. It will parse the page's Markdown and add injections points into
        the page's Markdown, which will be replaced by the renderer with the
        actual listing later on.

        Note that this method is intended to be called with the page during the
        `on_page_markdown` event, as it modifies the page's Markdown. Moreover,
        the Markdown must be explicitly passed, as we could otherwise run into
        inconsistencies when other plugins modify the Markdown.

        Arguments:
            page: The page.
            markdown: The page's Markdown.

        Returns:
            The page's Markdown with injection points.
        """
        assert isinstance(markdown, str)

        # Replace callback
        def replace(match: Match) -> str:
            config = self._resolve(page, match.group(2))

            # Compute listing identifier - as the author might include multiple
            # listings on a single page, we must make sure that the identifier
            # is unique, so we use the page source file path and the position
            # of the match within the page as an identifier.
            id = f"{page.file.src_uri}:{match.start()}-{match.end()}"

            # Replace whitespaces in the identifier that we computed, or we
            # can't just prefix it with "#" - see https://t.ly/U_hfp
            id = id.replace(" ", "-")
            self.data.add(Listing(page, id, config))

            # Replace directive with hx headline if listings are enabled, or
            # remove the listing entirely from the page and table of contents
            if self.config.listings:
                return "#" * self.depth + f" {id}/name {{ #{id}/slug }}"
            else:
                return

        # Hack: replace directive with an hx headline to mark the injection
        # point for the anchor links we will generate after parsing all pages.
        # By using an hx headline, we can make sure that the injection point
        # will always be a child of the preceding headline.
        directive = self.config.listings_directive
        return re.sub(
            r"(<!--\s*?{directive}(.*?)\s*-->)".format(directive = directive),
            replace, markdown, flags = re.I | re.M | re.S
        )

    def closest(self, mapping: Mapping) -> list[Listing]:
        """
        Get listings for the mapping ordered by closeness.

        Listings are sorted by closeness to the given page, i.e. the number of
        common path components. This is useful for hierarchical listings, where
        the tags of a page point to the closest listing featuring the tag, with
        the option to show all listings featuring that tag.

        Arguments:
            mapping: The mapping.

        Returns:
            The listings.
        """

        # Retrieve listings featuring tags of mapping
        listings: list[Listing] = []
        for listing in self.data:
            if any(listing & mapping):
                listings.append(listing)

        # Ranking callback
        def rank(listing: Listing) -> int:
            path = posixpath.commonpath([mapping.item.url, listing.page.url])
            return len(path)

        # Return listings ordered by closeness to mapping
        return sorted(listings, key = rank, reverse = True)

    def populate(
        self, listing: Listing, mappings: Iterable[Mapping], renderer: Renderer
    ) -> None:
        """
        Populate listing with tags featured in the mappings.

        Arguments:
            listing: The listing.
            mappings: The mappings.
            renderer: The renderer.
        """
        page = listing.page
        assert isinstance(page.content, str)

        # Add mappings to listing, passing shadow tags configuration
        for mapping in mappings:
            listing.add(mapping, hidden = listing.config.shadow)

        # Sort listings and tags - we can only do this after all mappings have
        # been added to the listing, because the tags inside the mappings do
        # not have a proper order yet, and we need to order them as specified
        # in the listing configuration.
        listing.tags = self._sort_listing_tags(listing.tags)

        # Render tags for listing headlines - the listing configuration allows
        # tp specify a custom layout, so we resolve the template for tags here
        name = posixpath.join(listing.config.layout, "tag.html")
        for tree in listing:
            tree.content = renderer.render(page, name, tag = tree.tag)

            # Sort mappings and subtrees of  listing tree
            tree.mappings = self._sort_listing(tree.mappings)
            tree.children = self._sort_listing_tags(tree.children)

        # Replace callback
        def replace(match: Match) -> str:
            hx = match.group()

            # Populate listing with anchor links to tags
            anchors = toc.populate(listing, self._slugify)
            if not anchors:
                return

            # Get reference to first tag in listing
            head = next(iter(anchors.values()))

            # Replace hx with actual level of listing and listing ids with
            # placeholders to create a format string for the headline
            hx = re.sub(
                r"<(/?)h{}\b".format(self.depth),
                r"<\g<1>h{}".format(head.level), hx
            )
            hx = re.sub(
                r"{id}\/(\w+)".format(id = listing.id),
                r"{\1}", hx, flags = re.I | re.M
            )

            # Render listing headlines
            for tree in listing:
                tree.content = hx.format(
                    slug = anchors[tree.tag].id,
                    name = tree.content
                )

            # Render listing - the listing configuration allows to specify a
            # custom layout, so we resolve the template for listings here
            name = posixpath.join(listing.config.layout, "listing.html")
            return "\n".join([
                renderer.render(page, name, listing = tree)
                    for tree in listing.tags.values()
            ])

        # Hack: replace hx headlines (injection points) we added when parsing
        # the page's Markdown with the actual listing content. Additionally,
        # replace anchor links in the table of contents with the hierarchy
        # generated from mapping over the listing, or remove them.
        page.content = re.sub(
            r"<h{x}[^>]+{id}.*?</h{x}>".format(
                id = f"{listing.id}/slug", x = self.depth
            ),
            replace, page.content, flags = re.I | re.M
        )

    def populate_all(
        self, mappings: Iterable[Mapping], renderer: Renderer
    ) -> None:
        """
        Populate all listings with tags featured in the mappings.

        This method is called by the tags plugin to populate all listings with
        the given mappings. It will also remove the injection points from the
        page's Markdown. Note that this method is intended to be called during
        the `on_env` event, after all pages have been rendered.

        Arguments:
            mappings: The mappings.
            renderer: The renderer.
        """
        for listing in self.data:
            self.populate(listing, mappings, renderer)

    # -------------------------------------------------------------------------

    def _resolve(self, page: Page, args: str) -> ListingConfig:
        """
        Resolve listing configuration.

        Arguments:
            page: The page the listing in embedded in.
            args: The arguments, as parsed from Markdown.

        Returns:
            The listing configuration.
        """
        data = yaml.safe_load(args)
        path = page.file.abs_src_path

        # Try to resolve available listing configuration
        if isinstance(data, str):
            config = self.config.listings_map.get(data, None)
            if not config:
                keys = ", ".join(self.config.listings_map.keys())
                raise PluginError(
                    f"Couldn't find listing configuration: {data}. Available "
                    f"configurations: {keys}"
                )

        # Otherwise, handle inline listing configuration
        else:
            config = ListingConfig(config_file_path = path)
            config.load_dict(data or {})

            # Validate listing configuration
            errors, warnings = config.validate()
            for _, w in warnings:
                path = os.path.relpath(path)
                log.warning(
                    f"Error reading listing configuration in '{path}':\n"
                    f"{w}"
                )
            for _, e in errors:
                path = os.path.relpath(path)
                raise PluginError(
                    f"Error reading listing configuration in '{path}':\n"
                    f"{e}"
                )

        # Inherit shadow tags configuration, unless explicitly set
        if not isinstance(config.shadow, bool):
            config.shadow = self.config.shadow

        # Inherit layout configuration, unless explicitly set
        if not isinstance(config.layout, str):
            config.layout = self.config.listings_layout

        # Inherit table of contents configuration, unless explicitly set
        if not isinstance(config.toc, bool):
            config.toc = self.config.listings_toc

        # Return listing configuration
        return config

    # -------------------------------------------------------------------------

    def _slugify(self, tag: Tag) -> str:
        """
        Slugify tag.

        If the tag hierarchy setting is enabled, the tag is expanded into a
        hierarchy of tags, all of which are then slugified and joined with the
        configured separator. Otherwise, the tag is slugified directly. This is
        necessary to keep the tag hierarchy in the slug.

        Arguments:
            tag: The tag.

        Returns:
            The slug.
        """
        slugify = self.config.tags_slugify
        tags = [tag.name]

        # Compute tag hierarchy, if configured
        hierarchy = self.config.tags_hierarchy_separator
        if self.config.tags_hierarchy:
            tags = tag.name.split(hierarchy)

        # Slugify tag hierarchy and join with separator
        separator = self.config.tags_slugify_separator
        return self.config.tags_slugify_format.format(
            slug = hierarchy.join(slugify(name, separator) for name in tags)
        )

    # -------------------------------------------------------------------------

    def _sort_listing(
        self, mappings: Iterable[Mapping]
    ) -> list[Mapping]:
        """
        Sort listing.

        When sorting a listing, we sort the mappings of the listing, which is
        why the caller must pass the mappings of the listing. That way, we can
        keep this implementation to be purely functional, without having to
        mutate the listing, which makes testing simpler.

        Arguments:
            mappings: The mappings.

        Returns:
            The sorted mappings.
        """
        return sorted(
            mappings,
            key = self.config.listings_sort_by,
            reverse = self.config.listings_sort_reverse
        )

    def _sort_listing_tags(
        self, children: dict[Tag, ListingTree]
    ) -> dict[Tag, ListingTree]:
        """
        Sort listing tags.

        When sorting a listing's tags, we sort the immediate subtrees of the
        listing, which is why the caller must pass the children of the listing.
        That way, we can keep this implementation to be purely functional,
        without having to mutate the listing.

        Arguments:
            children: The listing trees, each of which associated with a tag.

        Returns:
            The sorted listing trees.
        """
        return dict(sorted(
            children.items(),
            key = lambda item: self.config.listings_tags_sort_by(*item),
            reverse = self.config.listings_tags_sort_reverse
        ))

    def _sort_tags(
        self, tags: Iterable[Tag]
    ) -> list[Tag]:
        """
        Sort tags.

        Arguments:
            tags: The tags.

        Returns:
            The sorted tags.
        """
        return sorted(
            tags,
            key = self.config.tags_sort_by,
            reverse = self.config.tags_sort_reverse
        )

# -----------------------------------------------------------------------------
# Functions
# -----------------------------------------------------------------------------

def _print(manager: ListingManager, indent: int = 0) -> str:
    """
    Return a printable representation of a listing manager.

    Arguments:
        manager: The listing manager.
        indent: The indentation level.

    Returns:
        Printable representation.
    """
    lines: list[str] = []
    lines.append(" " * indent + f"ListingManager()")

    # Print listings
    for listing in manager:
        lines.append(" " * (indent + 2) + repr(listing))

    # Concatenate everything
    return "\n".join(lines)

# -----------------------------------------------------------------------------
# Data
# -----------------------------------------------------------------------------

# Set up logging
log = logging.getLogger("mkdocs.material.plugins.tags")



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

from material.plugins.tags.structure.tag import Tag
from mkdocs.structure.nav import Link

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

class TagReference(Tag):
    """
    A tag reference.

    Tag references are a subclass of tags that can have associated links, which
    is primarily used for linking tags to listings. The first link is used as
    the canonical link, which by default points to the closest listing that
    features the tag. This is considered to be the canonical listing.
    """

    def __init__(self, tag: Tag, links: list[Link] | None = None):
        """
        Initialize the tag reference.

        Arguments:
            tag: The tag.
            links: The links associated with the tag.
        """
        super().__init__(**vars(tag))
        self.links = links or []

    def __repr__(self) -> str:
        """
        Return a printable representation of the tag reference.

        Returns:
            Printable representation.
        """
        return f"TagReference('{self.name}')"

    # -------------------------------------------------------------------------

    links: list[Link]
    """
    The links associated with the tag.
    """

    # -------------------------------------------------------------------------

    @property
    def url(self) -> str | None:
        """
        Return the URL of the tag reference.

        Returns:
            The URL of the tag reference.
        """
        if self.links:
            return self.links[0].url
        else:
            return None



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import getpass
import glob
import json
import logging
import os
import platform
import re
import requests
import site
import sys
import yaml

from colorama import Fore, Style
from importlib.metadata import distributions, version
from io import BytesIO
from markdown.extensions.toc import slugify
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin, event_priority
from mkdocs.utils.yaml import get_yaml_loader
from zipfile import ZipFile, ZIP_DEFLATED

from .config import InfoConfig
from .patterns import get_exclusion_patterns

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Info plugin
class InfoPlugin(BasePlugin[InfoConfig]):

    # Initialize plugin
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Initialize incremental builds
        self.is_serve = False

        # Initialize empty members
        self.exclusion_patterns = []
        self.excluded_entries = []

    # Determine whether we're serving the site
    def on_startup(self, *, command, dirty):
        self.is_serve = command == "serve"

    # Create a self-contained example (run earliest) - determine all files that
    # are visible to MkDocs and are used to build the site, create an archive
    # that contains all of them, and print a summary of the archive contents.
    # The author must attach this archive to the bug report.
    @event_priority(100)
    def on_config(self, config):
        if not self.config.enabled:
            return

        # By default, the plugin is disabled when the documentation is served,
        # but not when it is built. This should nicely align with the expected
        # user experience when creating reproductions.
        if not self.config.enabled_on_serve and self.is_serve:
            return

        # Resolve latest version
        url = "https://github.com/squidfunk/mkdocs-material/releases/latest"
        res = requests.get(url, allow_redirects = False)

        # Check if we're running the latest version
        _, current = res.headers.get("location").rsplit("/", 1)
        present = version("mkdocs-material")
        if not present.startswith(current):
            log.error("Please upgrade to the latest version.")
            self._help_on_versions_and_exit(present, current)

        # Exit if archive creation is disabled
        if not self.config.archive:
            sys.exit(1)

        # Print message that we're creating a bug report
        log.info("Started archive creation for bug report")

        # Check that there are no overrides in place - we need to use a little
        # hack to detect whether the custom_dir setting was used without parsing
        # mkdocs.yml again - we check at which position the directory provided
        # by the theme resides, and if it's not the first one, abort.
        if config.theme.custom_dir:
            log.error("Please remove 'custom_dir' setting.")
            self._help_on_customizations_and_exit()

        # Check that there are no hooks in place - hooks can alter the behavior
        # of MkDocs in unpredictable ways, which is why they must be considered
        # being customizations. Thus, we can't offer support for debugging and
        # must abort here.
        if config.hooks:
            log.error("Please remove 'hooks' setting.")
            self._help_on_customizations_and_exit()

        # Assure all paths that will be validated are absolute. Convert possible
        # relative config_file_path to absolute. Its absolute directory path is
        # being later used to resolve other paths.
        config.config_file_path = _convert_to_abs(config.config_file_path)
        config_file_parent = os.path.dirname(config.config_file_path)

        # Convert relative custom_dir path to absolute. The Theme.custom_dir
        # property cannot be set, therefore a helper variable is used.
        if config.theme.custom_dir:
            abs_custom_dir = _convert_to_abs(
                config.theme.custom_dir,
                abs_prefix = config_file_parent
            )
        else:
            abs_custom_dir = ""

        # Extract the absolute path to projects plugin's directory to explicitly
        # support path validation and dynamic exclusion for the plugin
        projects_plugin = config.plugins.get("material/projects")
        if projects_plugin:
            abs_projects_dir = _convert_to_abs(
                projects_plugin.config.projects_dir,
                abs_prefix = config_file_parent
            )
        else:
            abs_projects_dir = ""

        # MkDocs removes the INHERIT configuration key during load, and doesn't
        # expose the information in any way, as the parent configuration is
        # merged into one. To validate that the INHERIT config file will be
        # included in the ZIP file the current config file must be loaded again
        # without parsing. Each file can have their own INHERIT key, so a list
        # of configurations is supported. The INHERIT path is converted during
        # load to absolute.
        loaded_configs = _load_yaml(config.config_file_path)
        if not isinstance(loaded_configs, list):
            loaded_configs = [loaded_configs]

        # It can happen that the drive letter case is inconsistent on Windows.
        # Therefore, assure first character to be uppercase for the following
        # checks. See: https://t.ly/9t1SU
        site_prefixes = list(map(capitalize, site.PREFIXES))
        cwd = capitalize(os.getcwd())

        # We need to make sure the user put every file in the current working
        # directory. To assure the reproduction inside the ZIP file can be run,
        # validate that the MkDocs paths are children of the current root.
        paths_to_validate = list(map(capitalize, [
            config.config_file_path,
            config.docs_dir,
            abs_custom_dir,
            abs_projects_dir,
            *[cfg.get("INHERIT", "") for cfg in loaded_configs]
        ]))

        # Convert relative hook paths to absolute path
        for hook in config.hooks:
            path = _convert_to_abs(hook, abs_prefix = config_file_parent)
            paths_to_validate.append(path)

        # Remove valid paths from the list
        for path in list(paths_to_validate):
            if not path or path.startswith(cwd):
                paths_to_validate.remove(path)

        # Report the invalid paths to the user
        if paths_to_validate:
            log.error("One or more paths aren't children of root")
            self._help_on_not_in_cwd(paths_to_validate)

        # Create in-memory archive and prompt author for a short descriptive
        # name for the archive, which is also used as the directory name. Note
        # that the name is slugified for better readability and stripped of any
        # file extension that the author might have entered.
        archive = BytesIO()
        example = input("\nPlease name your bug report (2-4 words): ")
        example, _ = os.path.splitext(example)
        example = "-".join([present, slugify(example, "-")])

        # Get local copy of the exclusion patterns
        self.exclusion_patterns = get_exclusion_patterns()
        self.excluded_entries = []

        # Exclude the site_dir at project root
        if capitalize(config.site_dir).startswith(cwd):
            self.exclusion_patterns.append(_resolve_pattern(config.site_dir))

        # Exclude the Virtual Environment directory. site.getsitepackages() has
        # inconsistent results across operating systems, and relies on the
        # PREFIXES that will contain the absolute path to the activated venv.
        for path in site_prefixes:
            if path.startswith(cwd):
                self.exclusion_patterns.append(_resolve_pattern(path))

        # Guess other Virtual Environment paths in case we forget to activate
        # them or in case we have multiple. Making sure which venv is activated
        # is not necessary, as it is an optional step in the guidelines.
        for abs_root, dirnames, filenames in os.walk(os.getcwd()):
            for filename in filenames:
                if filename.lower() != "pyvenv.cfg":
                    continue

                path = capitalize(abs_root)

                if path not in site_prefixes:
                    print(f"Possible inactive venv: {path}")
                    self.exclusion_patterns.append(_resolve_pattern(path))

        # Exclude site_dir for projects
        if projects_plugin:
            for path in glob.iglob(
                pathname = projects_plugin.config.projects_config_files,
                root_dir = abs_projects_dir,
                recursive = True
            ):
                current_config_file = os.path.join(abs_projects_dir, path)
                project_config = _get_project_config(current_config_file)
                pattern = _resolve_pattern(project_config.site_dir)
                self.exclusion_patterns.append(pattern)

        # Track dotpath inclusion to inform about it later
        contains_dotpath: bool = False

        # Create self-contained example from project
        files: list[str] = []
        with ZipFile(archive, "a", ZIP_DEFLATED, False) as f:
            for abs_root, dirnames, filenames in os.walk(os.getcwd()):
                # Set and print progress indicator
                indicator = f"Processing: {abs_root}"
                print(indicator, end="\r", flush=True)

                # Prune the folders in-place to prevent their processing
                for name in list(dirnames):
                    # Resolve the absolute directory path
                    path = os.path.join(abs_root, name)

                    # Exclude the directory and all subdirectories
                    if self._is_excluded(path):
                        dirnames.remove(name)
                        continue

                    # Warn about .dotdirectories
                    if _is_dotpath(path, log_warning = True):
                        contains_dotpath = True

                # Write files to the in-memory archive
                for name in filenames:
                    # Resolve the absolute file path
                    path = os.path.join(abs_root, name)

                    # Exclude the file
                    if self._is_excluded(path):
                        continue

                    # Warn about .dotfiles
                    if _is_dotpath(path, log_warning = True):
                        contains_dotpath = True

                    # Resolve the relative path to create a matching structure
                    path = os.path.relpath(path, os.path.curdir)
                    f.write(path, os.path.join(example, path))

                # Clear the line for the next indicator
                print(" " * len(indicator), end="\r", flush=True)

            # Add information on installed packages
            f.writestr(
                os.path.join(example, "requirements.lock.txt"),
                "\n".join(sorted([
                    "==".join([package.name, package.version])
                        for package in distributions()
                ]))
            )

            # Try to get login to replace it with USERNAME placeholder
            try:
                username = getpass.getuser()
            except Exception:
                username = "USERNAME"

            # Add information on platform
            f.writestr(
                os.path.join(example, "platform.json"),
                json.dumps(
                    {
                        "system": platform.platform(),
                        "architecture": platform.architecture(),
                        "python": platform.python_version(),
                        "cwd": os.getcwd(),
                        "command": " ".join([
                            sys.argv[0].rsplit(os.sep, 1)[-1],
                            *sys.argv[1:]
                        ]),
                        "env:$PYTHONPATH": os.getenv("PYTHONPATH", ""),
                        "env:$VIRTUAL_ENV": os.getenv("VIRTUAL_ENV", ""),
                        "sys.path": sys.path,
                        "excluded_entries": self.excluded_entries
                    },
                    default = str,
                    indent = 2
                ).replace(username, "USERNAME")
            )

            # Retrieve list of processed files
            for a in f.filelist:
                # Highlight .dotpaths in a more explicit manner
                color = (Fore.LIGHTYELLOW_EX if "/." in a.filename
                         else Fore.LIGHTBLACK_EX)
                files.append("".join([
                    color, a.filename, " ",
                    _size(a.compress_size)
                ]))

        # Finally, write archive to disk
        buffer = archive.getbuffer()
        with open(f"{example}.zip", "wb") as f:
            f.write(archive.getvalue())

        # Print summary
        log.info("Archive successfully created:")
        print(Style.NORMAL)

        # Print archive file names
        files.sort()
        for file in files:
            print(f"  {file}")

        # Print archive name
        print(Style.RESET_ALL)
        print("".join([
            "  ", f.name, " ",
            _size(buffer.nbytes, 10)
        ]))

        # Print warning when file size is excessively large
        print(Style.RESET_ALL)
        if buffer.nbytes > 1000000:
            log.warning("Archive exceeds recommended maximum size of 1 MB")

        # Print warning when file contains hidden .dotpaths
        if contains_dotpath:
            log.warning(
                "Archive contains dotpaths, which could contain sensitive "
                "information.\nPlease review them at the bottom of the list "
                "and share only necessary data to reproduce the issue."
            )

        # Aaaaaand done
        sys.exit(1)

    # -------------------------------------------------------------------------

    # Print help on versions and exit
    def _help_on_versions_and_exit(self, have, need):
        print(Fore.RED)
        print("  When reporting issues, please first upgrade to the latest")
        print("  version of Material for MkDocs, as the problem might already")
        print("  be fixed in the latest version. This helps reduce duplicate")
        print("  efforts and saves us maintainers time.")
        print(Style.NORMAL)
        print(f"  Please update from {have} to {need}.")
        print(Style.RESET_ALL)
        print(f"  pip install --upgrade --force-reinstall mkdocs-material")
        print(Style.NORMAL)

        # Exit, unless explicitly told not to
        if self.config.archive_stop_on_violation:
            sys.exit(1)

    # Print help on customizations and exit
    def _help_on_customizations_and_exit(self):
        print(Fore.RED)
        print("  When reporting issues, you must remove all customizations")
        print("  and check if the problem persists. If not, the problem is")
        print("  caused by your overrides. Please understand that we can't")
        print("  help you debug your customizations. Please remove:")
        print(Style.NORMAL)
        print("  - theme.custom_dir")
        print("  - hooks")
        print(Fore.YELLOW)
        print("  Additionally, please remove all third-party JavaScript or")
        print("  CSS not explicitly mentioned in our documentation:")
        print(Style.NORMAL)
        print("  - extra_css")
        print("  - extra_javascript")
        print(Fore.YELLOW)
        print("  If you're using customizations from the theme's documentation")
        print("  and you want to report a bug specific to those customizations")
        print("  then set the 'archive_stop_on_violation: false' option in the")
        print("  info plugin config.")
        print(Style.RESET_ALL)

        # Exit, unless explicitly told not to
        if self.config.archive_stop_on_violation:
            sys.exit(1)

    # Print help on not in current working directory and exit
    def _help_on_not_in_cwd(self, outside_root):
        print(Fore.RED)
        print("  The current working (root) directory:\n")
        print(f"    {os.getcwd()}\n")
        print("  is not a parent of the following paths:")
        print(Style.NORMAL)
        for path in outside_root:
            print(f"    {path}")
        print("\n  To assure that all project files are found please adjust")
        print("  your config or file structure and put everything within the")
        print("  root directory of the project.")
        print("\n  Please also make sure `mkdocs build` is run in the actual")
        print("  root directory of the project.")
        print(Style.RESET_ALL)

        # Exit, unless explicitly told not to
        if self.config.archive_stop_on_violation:
            sys.exit(1)

    # Check if path is excluded and should be omitted from the zip. Use pattern
    # matching for files and folders, and lookahead specific files in folders to
    # skip them. Side effect: Save excluded paths to save them in the zip file.
    def _is_excluded(self, abspath: str) -> bool:

        # Resolve the path into POSIX format to match the patterns
        pattern_path = _resolve_pattern(abspath, return_path = True)
        for pattern in self.exclusion_patterns:
            if re.search(pattern, pattern_path):
                log.debug(f"Excluded pattern '{pattern}': {abspath}")
                self.excluded_entries.append(f"{pattern} - {pattern_path}")
                return True

        # File exclusion should be limited to pattern matching
        if os.path.isfile(abspath):
            return False

        # Projects, which don't use the projects plugin for multi-language
        # support could have separate build folders for each config file or
        # language. Therefore, we exclude them with the assumption a site_dir
        # contains the sitemap file. Example of such a setup: https://t.ly/DLQcy
        sitemap_gz = os.path.join(abspath, "sitemap.xml.gz")
        if os.path.exists(sitemap_gz):
            log.debug(f"Excluded site_dir: {abspath}")
            self.excluded_entries.append(f"sitemap.xml.gz - {pattern_path}")
            return True

        return False

# -----------------------------------------------------------------------------
# Helper functions
# -----------------------------------------------------------------------------

# Print human-readable size
def _size(value, factor = 1):
    color = Fore.GREEN
    if   value > 100000 * factor: color = Fore.RED
    elif value >  25000 * factor: color = Fore.YELLOW
    for unit in ["B", "kB", "MB", "GB", "TB", "PB", "EB", "ZB"]:
        if abs(value) < 1000.0:
            return f"{color}{value:3.1f} {unit}"
        value /= 1000.0

# Get the absolute path with set prefix. To validate if a file is inside the
# current working directory it needs to be absolute, so that it is possible to
# check the prefix.
def _convert_to_abs(path: str, abs_prefix: str = None) -> str:
    if os.path.isabs(path): return path
    if abs_prefix is None: abs_prefix = os.getcwd()
    return os.path.normpath(os.path.join(abs_prefix, path))

# Get the loaded config, or a list with all loaded configs. MkDocs removes the
# INHERIT configuration key during load, and doesn't expose the information in
# any way, as the parent configuration is merged into one. The INHERIT path is
# needed for validation. This custom YAML loader replicates MkDocs' loading
# logic. Side effect: It converts the INHERIT path to absolute.
def _load_yaml(abs_src_path: str):

    with open(abs_src_path, encoding ="utf-8-sig") as file:
        source = file.read()

    try:
        result = yaml.load(source, Loader = get_yaml_loader()) or {}
    except yaml.YAMLError:
        result = {}

    if "INHERIT" in result:
        relpath = result.get('INHERIT')
        parent_path = os.path.dirname(abs_src_path)
        abspath = _convert_to_abs(relpath, abs_prefix = parent_path)
        if os.path.exists(abspath):
            result["INHERIT"] = abspath
            log.debug(f"Loading inherited configuration file: {abspath}")
            parent = _load_yaml(abspath)
            if isinstance(parent, list):
                result = [result, *parent]
            elif isinstance(parent, dict):
                result = [result, parent]

    return result

# Get a normalized POSIX path for the pattern matching with removed current
# working directory prefix. Directory paths end with a '/' to allow more control
# in the pattern creation for files and directories. The patterns are matched
# using the search function, so they are prefixed with ^ for specificity.
def _resolve_pattern(abspath: str, return_path: bool = False):
    path = capitalize(abspath).replace(capitalize(os.getcwd()), "", 1)
    path = path.replace(os.sep, "/").rstrip("/")

    if not path:
        return "/"

    # Check abspath, as the file needs to exist
    if not os.path.isfile(abspath):
        path = path + "/"

    return path if return_path else f"^{path}"

# Get project configuration with resolved absolute paths for validation
def _get_project_config(project_config_file: str):
    with open(project_config_file, encoding="utf-8-sig") as file:
        config = MkDocsConfig(config_file_path = project_config_file)
        config.load_file(file)

        # MkDocs transforms site_dir to absolute path during validation
        config.validate()

        return config

# Check if the path is a .dotpath. A warning can also be issued when the param
# is set. The function also returns a boolean to track results outside it.
def _is_dotpath(path: str, log_warning: bool = False) -> bool:
    posix_path = _resolve_pattern(path, return_path = True)
    name = posix_path.rstrip("/").rsplit("/", 1)[-1]
    if name.startswith("."):
        if log_warning:
            log.warning(f"The following .dotpath will be included: {path}")
        return True
    return False

# It can happen that the drive letter case is inconsistent on Windows.
# Capitalize the first character keeping the rest the same for comparison.
# See: https://t.ly/9t1SU
def capitalize(path: str):
    return path[0].upper() + path[1:] if path else path

# -----------------------------------------------------------------------------
# Data
# -----------------------------------------------------------------------------

# Set up logging
log = logging.getLogger("mkdocs.material.info")



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

import logging

from collections.abc import Callable
from mkdocs.config.config_options import Plugins
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.exceptions import PluginError
from mkdocs.plugins import BasePlugin, event_priority

from .config import GroupConfig

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Group plugin
class GroupPlugin(BasePlugin[GroupConfig]):
    supports_multiple_instances = True

    # Initialize plugin
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Initialize object attributes
        self.is_serve = False
        self.is_dirty = False

    # Determine whether we're serving the site
    def on_startup(self, *, command, dirty):
        self.is_serve = command == "serve"
        self.is_dirty = dirty

    # If the group is enabled, conditionally load plugins - at first, this might
    # sound easier than it actually is, as we need to jump through some hoops to
    # ensure correct ordering among plugins. We're effectively initializing the
    # plugins that are part of the group after all MkDocs finished initializing
    # all other plugins, so we need to patch the order of the methods. Moreover,
    # we must use MkDocs existing plugin collection, or we might have collisions
    # with other plugins that are not part of the group. As so often, this is a
    # little hacky, but has huge potential making plugin configuration easier.
    # There's one little caveat: the `__init__` and `on_startup` methods of the
    # plugins that are part of the group are called after all other plugins, so
    # the `event_priority` decorator for `on_startup` methods is effectively
    # useless. However, the `on_startup` method is only intended to set up the
    # plugin and doesn't receive anything else than the invoked command and
    # whether we're running a dirty build, so there should be no problems.
    @event_priority(150)
    def on_config(self, config):
        if not self.config.enabled:
            return

        # Retrieve plugin collection from configuration
        option: Plugins = dict(config._schema)["plugins"]
        assert isinstance(option, Plugins)

        # Load all plugins in group
        self.plugins: dict[str, BasePlugin] = {}
        try:
            for name, plugin in self._load(option):
                self.plugins[name] = plugin

        # The plugin could not be loaded, likely because it's not installed or
        # misconfigured, so we raise a plugin error for a nicer error message
        except Exception as e:
            raise PluginError(str(e))

        # Patch order of plugin methods
        for events in option.plugins.events.values():
            self._patch(events, config)

        # Invoke `on_startup` event for plugins in group
        command = "serve" if self.is_serve else "build"
        for method in option.plugins.events["startup"]:
            plugin = self._get_plugin(method)

            # Ensure that we have a method bound to a plugin (and not a hook)
            if plugin and plugin in self.plugins.values():
                method(command = command, dirty = self.is_dirty)

    # -------------------------------------------------------------------------

    # Retrieve plugin instance for bound method or nothing
    def _get_plugin(self, method: Callable):
        return getattr(method, "__self__", None)

    # Retrieve priority of plugin method
    def _get_priority(self, method: Callable):
        return getattr(method, "mkdocs_priority", 0)

    # Retrieve position of plugin
    def _get_position(self, plugin: BasePlugin, config: MkDocsConfig) -> int:
        for at, (_, candidate) in enumerate(config.plugins.items()):
            if plugin == candidate:
                return at

    # -------------------------------------------------------------------------

    # Load plugins that are part of the group
    def _load(self, option: Plugins):
        for name, data in option._parse_configs(self.config.plugins):
            yield option.load_plugin_with_namespace(name, data)

    # -------------------------------------------------------------------------

    # Patch order of plugin methods - all other plugin methods are already in
    # the right order, so we only need to check those that are part of the group
    # and bubble them up into the right location. Some plugin methods may define
    # priorities, so we need to make sure to order correctly within those.
    def _patch(self, methods: list[Callable], config: MkDocsConfig):
        position = self._get_position(self, config)
        for at in reversed(range(1, len(methods))):
            tail = methods[at - 1]
            head = methods[at]

            # Skip if the plugin is not part of the group
            plugin = self._get_plugin(head)
            if not plugin or plugin not in self.plugins.values():
                continue

            # Skip if the previous method has a higher priority than the current
            # one, because we know we can't swap them anyway
            if self._get_priority(tail) > self._get_priority(head):
                continue

            # Ensure that we have a method bound to a plugin (and not a hook)
            plugin = self._get_plugin(tail)
            if not plugin:
                continue

            # Both methods have the same priority, so we check if the ordering
            # of both methods is violated, and if it is, swap them
            if (position < self._get_position(plugin, config)):
                methods[at], methods[at - 1] = tail, head

# -----------------------------------------------------------------------------
# Data
# -----------------------------------------------------------------------------

# Set up logging
log = logging.getLogger("mkdocs.material.group")



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from datetime import date, datetime, time, timezone
from mkdocs.config.base import BaseConfigOption, Config, ValidationError
from mkdocs.config.config_options import ListOfItems, T
from mkdocs.structure.files import Files
from mkdocs.structure.nav import (
    Navigation, _add_parent_links, _data_to_navigation
)
from typing import Dict

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Date dictionary
class DateDict(Dict[str, datetime]):

    # Initialize date dictionary
    def __init__(self, data: dict):
        super().__init__(data)

        # Ensure presence of `date.created`
        self.created: datetime = data["created"]

    # Allow attribute access
    def __getattr__(self, name: str):
        if name in self:
            return self[name]

# -----------------------------------------------------------------------------

# Post date option
class PostDate(BaseConfigOption[DateDict]):

    # Initialize post dates
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # Normalize the supported types for post dates to datetime
    def pre_validation(self, config: Config, key_name: str):

        # If the date points to a scalar value, convert it to a dictionary, as
        # we want to allow the author to specify custom and arbitrary dates for
        # posts. Currently, only the `created` date is mandatory, because it's
        # needed to sort posts for views.
        if not isinstance(config[key_name], dict):
            config[key_name] = { "created": config[key_name] }

        # Convert all date values to datetime
        for key, value in config[key_name].items():

            # Handle datetime - since datetime is a subclass of date, we need
            # to check it first, or we lose the time - see https://t.ly/-KG9N
            if isinstance(value, datetime):
                # Set timezone to UTC if not set
                if value.tzinfo is None:
                    config[key_name][key] = value.replace(tzinfo=timezone.utc)
                continue;


            # Handle date - we set 00:00:00 as the default time, if the author
            # only supplied a date, and convert it to datetime in UTC
            if isinstance(value, date):
                config[key_name][key] = datetime.combine(value, time()).replace(tzinfo=timezone.utc)

        # Initialize date dictionary
        config[key_name] = DateDict(config[key_name])

    # Ensure each date value is of type datetime
    def run_validation(self, value: DateDict):
        for key in value:
            if not isinstance(value[key], datetime):
                raise ValidationError(
                    f"Expected type: {date} or {datetime} "
                    f"but received: {type(value[key])}"
                )

        # Ensure presence of `date.created`
        if not value.created:
            raise ValidationError(
                "Expected 'created' date when using dictionary syntax"
            )

        # Return date dictionary
        return value

# -----------------------------------------------------------------------------

# Post links option
class PostLinks(BaseConfigOption[Navigation]):

    # Create navigation from structured items - we don't need to provide a
    # configuration object to the function, because it will not be used
    def run_validation(self, value: object):
        items = _data_to_navigation(value, Files([]), None)
        _add_parent_links(items)

        # Return navigation
        return Navigation(items, [])

# -----------------------------------------------------------------------------

# Unique list of items
class UniqueListOfItems(ListOfItems[T]):

    # Ensure that each item is unique
    def run_validation(self, value: object):
        data = super().run_validation(value)
        return list(dict.fromkeys(data))



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

import logging

from material.utilities.filter import FileFilter, FilterConfig
from mkdocs.structure.pages import _RelativePathTreeprocessor
from markdown import Extension, Markdown
from markdown.treeprocessors import Treeprocessor
from mkdocs.exceptions import ConfigurationError
from urllib.parse import urlparse
from xml.etree.ElementTree import Element

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

class PreviewProcessor(Treeprocessor):
    """
    A Markdown treeprocessor to enable instant previews on links.

    Note that this treeprocessor is dependent on the `relpath` treeprocessor
    registered programmatically by MkDocs before rendering a page.
    """

    def __init__(self, md: Markdown, config: dict):
        """
        Initialize the treeprocessor.

        Arguments:
            md: The Markdown instance.
            config: The configuration.
        """
        super().__init__(md)
        self.config = config

    def run(self, root: Element):
        """
        Run the treeprocessor.

        Arguments:
            root: The root element of the parsed Markdown document.
        """
        at = self.md.treeprocessors.get_index_for_name("relpath")

        # Hack: Python Markdown has no notion of where it is, i.e., which file
        # is being processed. This seems to be a deliberate design decision, as
        # it is not possible to access the file path of the current page, but
        # it might also be an oversight that is now impossible to fix. However,
        # since this extension is only useful in the context of Material for
        # MkDocs, we can assume that the _RelativePathTreeprocessor is always
        # present, telling us the file path of the current page. If that ever
        # changes, we would need to wrap this extension in a plugin, but for
        # the time being we are sneaky and will probably get away with it.
        processor = self.md.treeprocessors[at]
        if not isinstance(processor, _RelativePathTreeprocessor):
            raise TypeError("Relative path processor not registered")

        # Normalize configurations
        configurations = self.config["configurations"]
        configurations.append({
            "sources": self.config.get("sources"),
            "targets": self.config.get("targets")
        })

        # Walk through all configurations - @todo refactor so that we don't
        # iterate multiple times over the same elements
        for configuration in configurations:

            # Skip, if the configuration defines nothing – we could also fix
            # this in the file filter, but we first fix it here and check if
            # it generalizes well enough to other inclusion/exclusion sites,
            # because here, it would hinder the ability to automaticaly
            # include all sources, while excluding specific targets.
            if (
                not configuration.get("sources") and
                not configuration.get("targets")
            ):
                continue

            # Skip if page should not be considered
            filter = get_filter(configuration, "sources")
            if not filter(processor.file):
                continue

            # Walk through all links and add preview attributes
            filter = get_filter(configuration, "targets")
            for el in root.iter("a"):
                href = el.get("href")
                if not href:
                    continue

                # Skip footnotes
                if "footnote-ref" in el.get("class", ""):
                    continue

                # Skip external links
                url = urlparse(href)
                if url.scheme or url.netloc:
                    continue

                # Add preview attribute to internal links
                for path in processor._possible_target_uris(
                    processor.file, url.path,
                    processor.config.use_directory_urls
                ):
                    target = processor.files.get_file_from_path(path)
                    if not target:
                        continue

                    # Include, if filter matches
                    if filter(target):
                        el.set("data-preview", "")

# -----------------------------------------------------------------------------

class PreviewExtension(Extension):
    """
    A Markdown extension to enable instant previews on links.

    This extensions allows to automatically add the `data-preview` attribute to
    internal links matching specific criteria, so Material for MkDocs renders a
    nice preview on hover as part of a tooltip. It is the recommended way to
    add previews to links in a programmatic way.
    """

    def __init__(self, *args, **kwargs):
        """
        """
        self.config = {
            "configurations": [[], "Filter configurations"],
            "sources": [{}, "Link sources"],
            "targets": [{}, "Link targets"]
        }
        super().__init__(*args, **kwargs)

    def extendMarkdown(self, md: Markdown):
        """
        Register Markdown extension.

        Arguments:
            md: The Markdown instance.
        """
        md.registerExtension(self)

        # Create and register treeprocessor - we use the same priority as the
        # `relpath` treeprocessor, the latter of which is guaranteed to run
        # after our treeprocessor, so we can check the original Markdown URIs
        # before they are resolved to URLs.
        processor = PreviewProcessor(md, self.getConfigs())
        md.treeprocessors.register(processor, "preview", 0)

# -----------------------------------------------------------------------------
# Functions
# -----------------------------------------------------------------------------

def get_filter(settings: dict, key: str):
    """
    Get file filter from settings.

    Arguments:
        settings: The settings.
        key: The key in the settings.

    Returns:
        The file filter.
    """
    config = FilterConfig()
    config.load_dict(settings.get(key) or {})

    # Validate filter configuration
    errors, warnings = config.validate()
    for _, w in warnings:
        log.warning(
            f"Error reading filter configuration in '{key}':\n"
            f"{w}"
        )
    for _, e in errors:
        raise ConfigurationError(
            f"Error reading filter configuration in '{key}':\n"
            f"{e}"
        )

    # Return file filter
    return FileFilter(config = config) # type: ignore

def makeExtension(**kwargs):
    """
    Register Markdown extension.

    Arguments:
        **kwargs: Configuration options.

    Returns:
        The Markdown extension.
    """
    return PreviewExtension(**kwargs)

# -----------------------------------------------------------------------------
# Data
# -----------------------------------------------------------------------------

# Set up logging
log = logging.getLogger("mkdocs.material.extensions.preview")



# Copyright (c) 2016-2025 Martin Donath <martin.donath@squidfunk.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import annotations

import os

from collections.abc import Callable
from material.plugins.projects.structure import Project
from mkdocs.livereload import LiveReloadServer

from .handler import ProjectChanged, ProjectAddedOrRemoved

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Projects watcher
class ProjectsWatcher:

    # Set of watched paths
    watched: set[str] = set()

    # Initialize projects watcher
    def __init__(self, server: LiveReloadServer):
        self.server = server

    # Watch project and invoke function on change
    def watch(self, project: Project, fn: Callable):
        self._on_project_changed(project, fn)
        self._on_project_added_or_removed(project, fn)

    # Stop watching project
    def unwatch(self, project: Project):

        # Traverse all watched paths
        root = os.path.dirname(project.config.config_file_path)
        for path in [*self.watched]:

            # Remove path from watched paths
            if path.startswith(root):
                self.server.unwatch(path)
                self.watched.remove(path)

    # -------------------------------------------------------------------------

    # Register event handler for project changes
    def _on_project_changed(self, project: Project, fn: Callable):

        # Resolve project root and docs directory
        root = os.path.dirname(project.config.config_file_path)
        docs = os.path.join(root, project.config.docs_dir)

        # Collect all paths to watch
        paths = set([docs, project.config.config_file_path])
        for path in project.config.watch:
            paths.add(os.path.join(root, path))

        # Register event handler for unwatched paths
        handler = ProjectChanged(project, fn)
        for path in paths - self.watched:
            self.server.watch(path)

            # Add path and its descendents to watched paths
            self.server.observer.schedule(handler, path, recursive = True)
            self.watched.add(path)

    # Register event handler for project additions and removals
    def _on_project_added_or_removed(self, project: Project, fn: Callable):

        # Resolve project root and path to projects directory
        root = os.path.dirname(project.config.config_file_path)
        path = os.path.join(root, project.plugin.projects_dir)

        # Register event handler for unwatched paths
        handler = ProjectAddedOrRemoved(project, fn)
        if path not in self.watched and os.path.isdir(path):

            # Add path to watched paths
            self.server.observer.schedule(handler, path)
            self.watched.add(path)
